"""
Module to provide processing for thematic break leaf blocks.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.container_blocks.container_helper import ContainerHelper
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.leaf_blocks.leaf_block_helper import LeafBlockHelper
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.stack_token import (
    ListStackToken,
    ParagraphStackToken,
    StackToken,
)
from pymarkdown.tokens.thematic_break_markdown_token import ThematicBreakMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class ThematicLeafBlockProcessor:
    """
    Module to provide processing for thematic break leaf blocks.
    """

    __thematic_break_characters = "*_-"

    @staticmethod
    def is_thematic_break(
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        skip_whitespace_check: bool = False,
        whitespace_allowed_between_characters: bool = True,
    ) -> Tuple[Optional[str], Optional[int]]:
        """
        Determine whether or not we have a thematic break.
        """

        thematic_break_character, end_of_break_index = None, None
        is_thematic_character = ParserHelper.is_character_at_index_one_of(
            line_to_parse,
            start_index,
            ThematicLeafBlockProcessor.__thematic_break_characters,
        )
        POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        POGGER.debug("skip_whitespace_check>>$", skip_whitespace_check)
        POGGER.debug("is_thematic_character>>$", is_thematic_character)
        if (
            TabHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
            or skip_whitespace_check
        ) and is_thematic_character:
            POGGER.debug("checking for thematic break")
            start_char, index, char_count, line_to_parse_size = (
                line_to_parse[start_index],
                start_index,
                0,
                len(line_to_parse),
            )

            while index < line_to_parse_size:
                if (
                    whitespace_allowed_between_characters
                    and ParserHelper.is_character_at_index_whitespace(
                        line_to_parse, index
                    )
                ):
                    index += 1
                elif line_to_parse[index] == start_char:
                    index += 1
                    char_count += 1
                else:
                    break  # pragma: no cover

            POGGER.debug("char_count>>$", char_count)
            POGGER.debug("index>>$", index)
            POGGER.debug("line_to_parse_size>>$", line_to_parse_size)
            if char_count >= 3 and index == line_to_parse_size:
                thematic_break_character, end_of_break_index = start_char, index

        return thematic_break_character, end_of_break_index

    @staticmethod
    def __handle_existing_paragraph_special(
        parser_state: ParserState,
        grab_bag: ContainerGrabBag,
        new_tokens: List[MarkdownToken],
    ) -> None:
        if (
            not parser_state.token_stack[-1].is_list
            or grab_bag.text_removed_by_container is None
        ):
            return
        stack_list_token = cast(ListStackToken, parser_state.token_stack[-1])
        indent_delta = stack_list_token.indent_level - len(
            grab_bag.text_removed_by_container
        )
        if indent_delta > 0:
            stack_index = len(parser_state.token_stack) - 1
            best_stack_index = -1
            while stack_index > 0 and parser_state.token_stack[stack_index].is_list:
                new_list_stack_token = cast(
                    ListStackToken, parser_state.token_stack[stack_index]
                )
                new_indent_delta = new_list_stack_token.indent_level - len(
                    grab_bag.text_removed_by_container
                )
                if new_indent_delta <= 0:
                    break
                best_stack_index = stack_index
                stack_index -= 1
            assert best_stack_index != -1
            closed_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                was_forced=True,
                include_lists=True,
                until_this_index=best_stack_index,
            )
            new_tokens.extend(closed_tokens)
            # if parser_state.token_stack[-1].is_list:
            #     list_token = cast(
            #         ListStartMarkdownToken,
            #         parser_state.token_stack[-1].matching_markdown_token,
            #     )
            #     assert ">" in grab_bag.text_removed_by_container
            #     bq_start_index = grab_bag.text_removed_by_container.rindex(">")
            #     assert bq_start_index != len(grab_bag.text_removed_by_container) - 1
            #     # real_indent_delta = len(grab_bag.text_removed_by_container) - (
            #     #     bq_start_index + 2
            #     # )
            #     # list_token.add_leading_spaces(" " * real_indent_delta)

    @staticmethod
    def __handle_existing_paragraph(
        parser_state: ParserState,
        grab_bag: ContainerGrabBag,
        new_tokens: List[MarkdownToken],
        block_quote_data: BlockQuoteData,
    ) -> List[MarkdownToken]:
        force_paragraph_close_if_present = (
            block_quote_data.current_count == 0 and block_quote_data.stack_count > 0
        )
        new_tokens, _ = parser_state.close_open_blocks_fn(
            parser_state,
            only_these_blocks=[ParagraphStackToken],
            was_forced=force_paragraph_close_if_present,
        )
        if new_tokens and grab_bag.text_removed_by_container:
            ThematicLeafBlockProcessor.__handle_existing_paragraph_special(
                parser_state, grab_bag, new_tokens
            )
        return new_tokens

    @staticmethod
    def __handle_special_case(
        parser_state: ParserState, new_tokens: List[MarkdownToken]
    ) -> None:
        if (
            new_tokens
            and new_tokens[-1].is_list_end
            and parser_state.token_stack[-1].is_block_quote
        ):
            stack_index = len(parser_state.token_stack) - 1
            while stack_index > 0 and not parser_state.token_stack[stack_index].is_list:
                stack_index -= 1
            if stack_index != 0:
                list_end_token = cast(EndMarkdownToken, new_tokens[-1])
                last_list_markdown_token = cast(
                    ListStartMarkdownToken, list_end_token.start_markdown_token
                )
                if last_list_markdown_token.leading_spaces is not None:
                    inner_list_markdown_token = cast(
                        ListStartMarkdownToken,
                        parser_state.token_stack[stack_index].matching_markdown_token,
                    )
                    POGGER.debug(
                        "__handle_special_case>>last_list_markdown_token>>$",
                        last_list_markdown_token,
                    )
                    leading_space_to_move = (
                        last_list_markdown_token.remove_last_leading_space()
                    )
                    POGGER.debug(
                        "__handle_special_case>>last_list_markdown_token>>$",
                        last_list_markdown_token,
                    )
                    assert leading_space_to_move is not None
                    POGGER.debug(
                        "__handle_special_case>>list_token>>$",
                        inner_list_markdown_token,
                    )
                    if leading_space_to_move:
                        leading_space_to_move += ParserLogger.blah_sequence
                    inner_list_markdown_token.add_leading_spaces(leading_space_to_move)
                    POGGER.debug(
                        "__handle_special_case>>list_token>>$",
                        inner_list_markdown_token,
                    )

    @staticmethod
    def parse_thematic_break(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        block_quote_data: BlockQuoteData,
        original_line: str,
        grab_bag: ContainerGrabBag,
    ) -> List[MarkdownToken]:
        """
        Handle the parsing of a thematic break.
        """

        ex_ws = LeafBlockHelper.realize_leading_whitespace(
            parser_state, position_marker, extracted_whitespace, original_line
        )
        new_tokens: List[MarkdownToken] = []
        start_char, index = ThematicLeafBlockProcessor.is_thematic_break(
            position_marker.text_to_parse,
            position_marker.index_number,
            ex_ws,
        )
        if start_char:
            old_top_of_stack = parser_state.token_stack[-1]

            POGGER.debug(
                "parse_thematic_break>>start",
            )
            if parser_state.token_stack[-1].is_paragraph:
                new_tokens = ThematicLeafBlockProcessor.__handle_existing_paragraph(
                    parser_state, grab_bag, new_tokens, block_quote_data
                )

            ThematicLeafBlockProcessor.__handle_special_case(parser_state, new_tokens)

            token_text = position_marker.text_to_parse[
                position_marker.index_number : index
            ]
            # POGGER.debug("extracted_whitespace>>:$:<", extracted_whitespace)
            # POGGER.debug("token_text>>:$:<", token_text)
            # POGGER.debug("original_line>>:$:<", original_line)
            split_tab, split_tab_with_block_quote_suffix = False, False
            extra_whitespace_prefix: Optional[str] = None
            if ParserHelper.tab_character in original_line:
                (
                    token_text,
                    split_tab,
                    split_tab_with_block_quote_suffix,
                    extra_whitespace_prefix,
                    extracted_whitespace,
                    _,
                ) = TabHelper.parse_thematic_break_with_tab(
                    original_line, token_text, extracted_whitespace
                )
                # POGGER.debug("extra_whitespace_prefix>>:$:<", extra_whitespace_prefix)
                # POGGER.debug("extracted_whitespace>>:$:<", extracted_whitespace)

            extracted_whitespace = ThematicLeafBlockProcessor.__perform_adjusts(
                parser_state,
                position_marker,
                extra_whitespace_prefix,
                extracted_whitespace,
                old_top_of_stack,
                new_tokens,
                start_char,
                token_text,
                block_quote_data,
                split_tab,
                split_tab_with_block_quote_suffix,
                grab_bag,
            )

            new_tokens.append(
                ThematicBreakMarkdownToken(
                    start_char,
                    extracted_whitespace,
                    token_text,
                    position_marker=position_marker,
                )
            )
        else:
            POGGER.debug(
                "parse_thematic_break>>not eligible",
            )
        return new_tokens

    @staticmethod
    def __perform_adjusts(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extra_whitespace_prefix: Optional[str],
        extracted_whitespace: str,
        old_top_of_stack: StackToken,
        new_tokens: List[MarkdownToken],
        start_char: str,
        token_text: str,
        block_quote_data: BlockQuoteData,
        split_tab: bool,
        split_tab_with_block_quote_suffix: bool,
        grab_bag: ContainerGrabBag,
    ) -> str:
        if split_tab and not split_tab_with_block_quote_suffix:
            ThematicLeafBlockProcessor.__parse_thematic_break_with_suffix(
                parser_state,
                position_marker,
                extra_whitespace_prefix,
                extracted_whitespace,
                old_top_of_stack,
                new_tokens,
                start_char,
                token_text,
                block_quote_data,
            )
        else:
            split_tab, extracted_whitespace, whitespace_prefix = (
                ContainerHelper.reduce_containers_if_required(
                    parser_state,
                    position_marker,
                    block_quote_data,
                    new_tokens,
                    split_tab,
                    extracted_whitespace,
                    grab_bag,
                )
            )
            if split_tab:
                TabHelper.adjust_block_quote_indent_for_tab(parser_state)
        return extracted_whitespace

    @staticmethod
    def __parse_thematic_break_with_suffix(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extra_whitespace_prefix: Optional[str],
        extracted_whitespace: str,
        old_top_of_stack: StackToken,
        new_tokens: List[MarkdownToken],
        start_char: str,
        token_text: str,
        block_quote_data: BlockQuoteData,
    ) -> None:
        POGGER.debug("parser_state.token_stack[-1]>>:$:<", parser_state.token_stack[-1])
        POGGER.debug("parser_state.token_stack>>:$:<", parser_state.token_stack)
        POGGER.debug("parser_state.token_document>>:$:<", parser_state.token_document)
        assert parser_state.token_stack[
            -1
        ].is_list, "This should only be a list container."
        modified_whitespace = (
            extra_whitespace_prefix + extracted_whitespace
            if extra_whitespace_prefix is not None
            else extracted_whitespace
        )
        TabHelper.adjust_block_quote_indent_for_tab(parser_state, modified_whitespace)
        POGGER.debug("parser_state.token_stack>>:$:<", parser_state.token_stack)
        POGGER.debug("parser_state.token_document>>:$:<", parser_state.token_document)
        LeafBlockHelper.correct_for_leaf_block_start_in_list(
            parser_state,
            position_marker.index_indent,
            old_top_of_stack,
            new_tokens,
            block_quote_data,
            was_token_already_added_to_stack=False,
            delay_tab_match=True,
        )
        POGGER.debug("parser_state.token_stack>>:$:<", parser_state.token_stack)
        POGGER.debug("parser_state.token_document>>:$:<", parser_state.token_document)
        POGGER.debug("new_tokens>>:$:<", new_tokens)

        POGGER.debug("start_char>>:$:<", start_char)
        POGGER.debug("extracted_whitespace>>:$:<", extracted_whitespace)
        POGGER.debug("token_text>>:$:<", token_text)



"""
Module to provide for linter instructions that can be embedded within the document.
"""

import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple

from typing_extensions import Protocol

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.extension_manager.extension_manager_constants import (
    ExtensionManagerConstants,
)
from pymarkdown.extension_manager.parser_extension import ParserExtension
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade
from pymarkdown.plugin_manager.found_plugin import FoundPlugin
from pymarkdown.tokens.markdown_token import MarkdownToken, MarkdownTokenClass
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


@dataclass
class GeneralPragmaDisableStart:
    """Data class to hold the start of a general pragma disable."""

    rule_id: str
    start_line: int


# pylint: disable=too-few-public-methods
class LogPragmaFailureProtocol(Protocol):
    """
    Protocol to specify a function that allows failures to be reported.
    """

    def __call__(  # noqa: E704
        self, scan_file: str, line_number: int, pragma_error: str
    ) -> None: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


class PragmaExtension(ParserExtension):
    """
    Extension to implement the pragma extensions.
    """

    def get_identifier(self) -> str:
        """
        Get the identifier associated with this extension.
        """
        return "linter-pragmas"

    def get_details(self) -> ExtensionDetails:
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id=self.get_identifier(),
            extension_name="Pragma Linter Instructions",
            extension_description="Allows parsing of instructions for the linter.",
            extension_enabled_by_default=True,
            extension_version="0.5.0",
            extension_interface_version=ExtensionManagerConstants.EXTENSION_INTERFACE_VERSION_BASIC,
            extension_url="https://pymarkdown.readthedocs.io/en/latest/extensions/pragmas/",
            extension_configuration=None,
        )

    def apply_configuration(
        self, extension_specific_facade: MyApplicationPropertiesFacade
    ) -> None:
        """
        Apply any configuration required by the extension.
        """
        _ = extension_specific_facade

    @staticmethod
    def look_for_pragmas(
        position_marker: PositionMarker,
        line_to_parse: str,
        container_depth: int,
        extracted_whitespace: Optional[str],
        parser_properties: ParseBlockPassProperties,
    ) -> bool:
        """
        Look for a pragma in the current line.
        """

        POGGER.debug("look_for_pragmas - >$<", line_to_parse)
        POGGER.debug("look_for_pragmas - ws >$<", extracted_whitespace)
        used_standard_prefix = line_to_parse.startswith(PragmaToken.pragma_prefix)
        used_alternate_prefix = line_to_parse.startswith(
            PragmaToken.pragma_alternate_prefix
        )
        if (
            not container_depth
            and not extracted_whitespace
            and (used_standard_prefix or used_alternate_prefix)
        ):
            was_extended_prefix = line_to_parse.startswith(
                PragmaToken.pragma_alternate_prefix
            )

            start_index, _ = ParserHelper.extract_spaces(
                line_to_parse,
                len(
                    PragmaToken.pragma_alternate_prefix
                    if was_extended_prefix
                    else PragmaToken.pragma_prefix
                ),
            )
            remaining_line = (
                line_to_parse[start_index:].rstrip(Constants.ascii_whitespace).lower()
            )
            if remaining_line.startswith(PragmaToken.pragma_title):
                found_matching_suffix = (
                    remaining_line.endswith(PragmaToken.pragma_suffix)
                    if used_standard_prefix
                    else remaining_line.endswith(PragmaToken.pragma_alternate_suffix)
                )
                if found_matching_suffix:
                    index_number = (
                        -position_marker.line_number
                        if was_extended_prefix
                        else position_marker.line_number
                    )
                    parser_properties.pragma_lines[index_number] = line_to_parse
                    POGGER.debug(
                        "pragma $ extracted - >$<", index_number, line_to_parse
                    )
                    return True
        POGGER.debug("pragma not extracted - >$<", line_to_parse)
        return False

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def compile_single_pragma(
        scan_file: str,
        next_line_number: int,
        pragma_lines: Dict[int, str],
        all_ids: Dict[str, FoundPlugin],
        document_pragmas: Dict[int, Set[str]],
        document_pragma_ranges: List[Tuple[int, int, Set[str]]],
        general_pragma_ranges: List[Tuple[int, int, str]],
        active_general_pragmas: Dict[str, GeneralPragmaDisableStart],
        log_pragma_failure: LogPragmaFailureProtocol,
        actual_pragma_line_numbers: List[int],
    ) -> None:
        """
        Compile a single pragma line, validating it before adding it to the dictionary of pragmas.
        """
        if next_line_number > 0:
            prefix_length = len(PragmaToken.pragma_prefix)
            actual_line_number = next_line_number
            suffix_to_use = PragmaToken.pragma_suffix
        else:
            prefix_length = len(PragmaToken.pragma_alternate_prefix)
            actual_line_number = -next_line_number
            suffix_to_use = PragmaToken.pragma_alternate_suffix

        actual_pragma_line_numbers.append(actual_line_number)

        line_after_prefix = pragma_lines[next_line_number][prefix_length:]
        after_whitespace_index, _ = ParserHelper.extract_spaces_verified(
            line_after_prefix, 0
        )
        after_whitespace_index, _ = ParserHelper.extract_spaces_verified(
            line_after_prefix, after_whitespace_index + len(PragmaToken.pragma_title)
        )
        command_data = line_after_prefix[after_whitespace_index : -len(suffix_to_use)]
        after_command_index, command = ParserHelper.extract_until_spaces_verified(
            command_data, 0
        )
        command = command.lower()
        if not command:
            log_pragma_failure(
                scan_file,
                actual_line_number,
                "Inline configuration specified without command.",
            )
        elif command == "disable":
            PragmaExtension.__handle_general_disable(
                command_data,
                after_command_index,
                scan_file,
                actual_line_number,
                log_pragma_failure,
                all_ids,
                command,
                active_general_pragmas,
            )
        elif command == "enable":
            PragmaExtension.__handle_general_enable(
                command_data,
                after_command_index,
                scan_file,
                actual_line_number,
                log_pragma_failure,
                all_ids,
                command,
                active_general_pragmas,
                general_pragma_ranges,
            )
        elif command == "disable-next-line":
            PragmaExtension.__handle_disable_next_line(
                command_data,
                after_command_index,
                log_pragma_failure,
                scan_file,
                actual_line_number,
                document_pragmas,
                all_ids,
                command,
            )
        elif command == "disable-num-lines":
            PragmaExtension.__handle_disable_num_lines(
                command_data,
                after_command_index,
                log_pragma_failure,
                scan_file,
                actual_line_number,
                document_pragma_ranges,
                all_ids,
                command,
            )
        else:
            log_pragma_failure(
                scan_file,
                actual_line_number,
                f"Inline configuration command '{command}' not understood.",
            )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def end(
        active_general_pragmas: Dict[str, GeneralPragmaDisableStart],
        general_pragma_ranges: List[Tuple[int, int, str]],
    ) -> None:
        """
        End the pragma extension compilation.
        """
        general_pragma_ranges.extend(
            (i.start_line, 9999, i.rule_id) for _, i in active_general_pragmas.items()
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_general_disable(
        command_data: str,
        after_command_index: int,
        scan_file: str,
        actual_line_number: int,
        log_pragma_failure: LogPragmaFailureProtocol,
        all_ids: Dict[str, FoundPlugin],
        command: str,
        active_general_pragmas: Dict[str, GeneralPragmaDisableStart],
    ) -> None:
        if processed_ids := PragmaExtension.__parse_rules_to_exclude(
            command_data,
            after_command_index,
            scan_file,
            actual_line_number,
            command,
            log_pragma_failure,
            all_ids,
        ):
            for next_rule_id in processed_ids:
                if next_rule_id in active_general_pragmas:
                    POGGER.debug(
                        "File $:line $: General pragma $ already disabled. Ignoring.",
                        scan_file,
                        actual_line_number,
                        next_rule_id,
                    )
                else:
                    active_general_pragmas[next_rule_id] = GeneralPragmaDisableStart(
                        next_rule_id, actual_line_number
                    )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_general_enable(
        command_data: str,
        after_command_index: int,
        scan_file: str,
        actual_line_number: int,
        log_pragma_failure: LogPragmaFailureProtocol,
        all_ids: Dict[str, FoundPlugin],
        command: str,
        active_general_pragmas: Dict[str, GeneralPragmaDisableStart],
        general_pragma_ranges: List[Tuple[int, int, str]],
    ) -> None:
        if processed_ids := PragmaExtension.__parse_rules_to_exclude(
            command_data,
            after_command_index,
            scan_file,
            actual_line_number,
            command,
            log_pragma_failure,
            all_ids,
        ):
            for next_rule_id in processed_ids:
                if next_rule_id in active_general_pragmas:
                    disable_record = active_general_pragmas[next_rule_id]
                    general_pragma_ranges.append(
                        (
                            disable_record.start_line,
                            actual_line_number,
                            disable_record.rule_id,
                        )
                    )
                    del active_general_pragmas[next_rule_id]
                else:
                    POGGER.debug(
                        "File $:line $: General pragma $ already enabled. Ignoring.",
                        scan_file,
                        actual_line_number,
                        next_rule_id,
                    )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_disable_next_line(
        command_data: str,
        after_command_index: int,
        log_pragma_failure: LogPragmaFailureProtocol,
        scan_file: str,
        actual_line_number: int,
        document_pragmas: Dict[int, Set[str]],
        all_ids: Dict[str, FoundPlugin],
        command: str,
    ) -> None:
        if processed_ids := PragmaExtension.__parse_rules_to_exclude(
            command_data,
            after_command_index,
            scan_file,
            actual_line_number,
            command,
            log_pragma_failure,
            all_ids,
        ):
            document_pragmas[actual_line_number + 1] = processed_ids

    @staticmethod
    def __parse_rules_to_exclude(
        command_data: str,
        after_number_index: int,
        scan_file: str,
        actual_line_number: int,
        command: str,
        log_pragma_failure: LogPragmaFailureProtocol,
        all_ids: Dict[str, FoundPlugin],
    ) -> Set[str]:
        ids_to_disable = command_data[after_number_index:].split(",")
        processed_ids = set()
        for next_id in ids_to_disable:
            next_id = next_id.strip(" ").lower()
            if not next_id:
                log_pragma_failure(
                    scan_file,
                    actual_line_number,
                    f"Inline configuration command '{command}' specified a plugin with a blank id.",
                )
            elif next_id in all_ids:
                normalized_id = all_ids[next_id].plugin_id
                processed_ids.add(normalized_id)
            else:
                log_pragma_failure(
                    scan_file,
                    actual_line_number,
                    f"Inline configuration command '{command}' unable to find a plugin with the id '{next_id}'.",
                )
        return processed_ids

    # pylint: enable=too-many-arguments
    # pylint: disable=too-many-arguments

    @staticmethod
    def __handle_disable_num_lines_parse(
        command_data: str,
        after_command_index: int,
        log_pragma_failure: LogPragmaFailureProtocol,
        scan_file: str,
        actual_line_number: int,
        command: str,
    ) -> Tuple[bool, int, Optional[int]]:
        after_space_index, _ = ParserHelper.extract_spaces_verified(
            command_data, after_command_index
        )
        if after_space_index == len(command_data):
            log_pragma_failure(
                scan_file,
                actual_line_number,
                f"Inline configuration command '{command}' was not followed by a count and a list of plugin ids to temporarily disable.",
            )
            return False, -1, None
        after_num_index, extracted_number = ParserHelper.extract_until_spaces_verified(
            command_data, after_space_index
        )

        try:
            count_value = int(extracted_number)
        except ValueError:
            count_value = -1
        if count_value < 1:
            log_pragma_failure(
                scan_file,
                actual_line_number,
                f"Inline configuration command '{command}' specified a count '{extracted_number}' that is not a valid positive integer.",
            )
            return False, -1, None

        after_number_index, _ = ParserHelper.extract_spaces_verified(
            command_data, after_num_index
        )
        if after_number_index == len(command_data):
            log_pragma_failure(
                scan_file,
                actual_line_number,
                f"Inline configuration command '{command}' and its count were not followed by a list of plugin ids to temporarily disable.",
            )
            return False, -1, None
        return True, after_number_index, count_value

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_disable_num_lines(
        command_data: str,
        after_command_index: int,
        log_pragma_failure: LogPragmaFailureProtocol,
        scan_file: str,
        actual_line_number: int,
        document_pragma_ranges: List[Tuple[int, int, Set[str]]],
        all_ids: Dict[str, FoundPlugin],
        command: str,
    ) -> None:
        (
            is_ok,
            after_number_index,
            count_value,
        ) = PragmaExtension.__handle_disable_num_lines_parse(
            command_data,
            after_command_index,
            log_pragma_failure,
            scan_file,
            actual_line_number,
            command,
        )
        if not is_ok:
            return

        if processed_ids := PragmaExtension.__parse_rules_to_exclude(
            command_data,
            after_number_index,
            scan_file,
            actual_line_number,
            command,
            log_pragma_failure,
            all_ids,
        ):
            assert count_value is not None
            pragma_tuple = (
                actual_line_number + 1,
                actual_line_number + count_value,
                processed_ids,
            )
            document_pragma_ranges.append(pragma_tuple)

    # pylint: enable=too-many-arguments


class PragmaToken(MarkdownToken):
    """
    Token that contains the pragmas for the document.
    """

    pragma_prefix = "<!--"
    pragma_alternate_prefix = "<!---"
    pragma_title = "pyml "
    pragma_suffix = "-->"
    pragma_alternate_suffix = "--->"

    def __init__(self, pragma_lines: Dict[int, str]) -> None:
        self.__pragma_lines = pragma_lines

        MarkdownToken.__init__(
            self,
            MarkdownToken._token_pragma,
            MarkdownTokenClass.SPECIAL,
            is_extension=True,
            extra_data="",
        )
        self.__compose_extra_data_field()

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        serialized_pragmas = "".join(
            f";{next_line_number}:{self.__pragma_lines[next_line_number]}"
            for next_line_number in self.__pragma_lines
        )
        self._set_extra_data(serialized_pragmas[1:])

    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_pragma

    @property
    def pragma_lines(self) -> Dict[int, str]:
        """
        Returns the pragma lines for the document.
        """
        return self.__pragma_lines

    def adjust_pragma_line_number(
        self, initial_line_number: int, new_line_number: int
    ) -> None:
        """Perform an adjustment to the line number of a given pragma."""
        old_pragma = self.__pragma_lines[initial_line_number]
        del self.__pragma_lines[initial_line_number]
        self.__pragma_lines[new_line_number] = old_pragma
        self.__compose_extra_data_field()

    def register_for_markdown_transform(
        self,
        registration_function: RegisterMarkdownTransformHandlersProtocol,
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """

        # Note, because the Pragma token contains every pragma contained
        # within the file, this is handled globally in TransformToMarkdown's
        # __handle_pragma_processing.
        _ = registration_function

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            PragmaToken,
            PragmaToken.__handle_pragma_token,
            None,
        )

    @staticmethod
    def __handle_pragma_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return output_html



from typing import Dict, List, Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.tokens.link_reference_definition_markdown_token import (
    LinkReferenceDefinitionMarkdownToken,
)
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class ContainerTokenManager:
    """
    TBD
    """

    def __init__(self) -> None:
        """
        TBD
        """
        self.container_token_stack: List[MarkdownToken] = []
        self.bq_line_index: Dict[int, int] = {}
        self.last_leaf_token: Optional[MarkdownToken] = None
        self.list_adjust_map: Dict[int, int] = {}

    def clear(self) -> None:
        """
        TBD
        """
        self.container_token_stack = []
        self.bq_line_index = {}
        self.last_leaf_token = None

    @classmethod
    def __is_simple_delta(cls, token: MarkdownToken) -> bool:
        return (
            token.is_blank_line
            or token.is_thematic_break
            or token.is_atx_heading
            or token.is_paragraph_end
            or token.is_inline_hard_break
        )

    @classmethod
    def __is_remember_leaf_token(cls, token: MarkdownToken) -> bool:
        return bool(
            token.is_setext_heading
            or token.is_indented_code_block
            or token.is_html_block
        )

    @classmethod
    def __is_clear_leaf_token(cls, token: MarkdownToken) -> bool:
        return bool(
            token.is_indented_code_block_end
            or token.is_html_block_end
            or token.is_setext_heading_end
            or token.is_fenced_code_block_end
        )

    def __manage_leaf_tokens_text(self, token: MarkdownToken) -> int:
        assert self.last_leaf_token is not None
        text_token = cast(TextMarkdownToken, token)
        if self.last_leaf_token.is_setext_heading:
            return (
                text_token.end_whitespace.count(ParserHelper.newline_character)
                if text_token.end_whitespace is not None
                else 0
            )
        assert self.last_leaf_token.is_html_block or self.last_leaf_token.is_code_block
        return text_token.token_text.count(ParserHelper.newline_character) + 1

    def __manage_leaf_tokens(self, token: MarkdownToken) -> None:
        bq_delta = 0
        if self.__is_simple_delta(token):
            bq_delta = 1
        elif self.__is_remember_leaf_token(token):
            self.last_leaf_token = token
        elif self.__is_clear_leaf_token(token):
            self.last_leaf_token = None
            if token.is_setext_heading_end or token.is_fenced_code_block_end:
                bq_delta = 1
        elif token.is_fenced_code_block:
            bq_delta = 1
            self.last_leaf_token = token
        elif token.is_paragraph:
            paragraph_token = cast(ParagraphMarkdownToken, token)
            bq_delta = paragraph_token.extracted_whitespace.count(
                ParserHelper.newline_character
            )
        elif token.is_link_reference_definition:
            bq_delta = self.__manage_lrd_token(token)
        elif token.is_text and self.last_leaf_token:
            bq_delta = self.__manage_leaf_tokens_text(token)
        self.bq_line_index[len(self.container_token_stack)] += bq_delta

    @classmethod
    def __manage_lrd_token(cls, token: MarkdownToken) -> int:
        lrd_token = cast(LinkReferenceDefinitionMarkdownToken, token)
        assert lrd_token.link_title_raw is not None
        assert lrd_token.link_title_whitespace is not None
        assert lrd_token.link_destination_whitespace is not None
        assert lrd_token.link_name_debug is not None
        return (
            1
            + lrd_token.link_name_debug.count(ParserHelper.newline_character)
            + lrd_token.link_destination_whitespace.count(
                ParserHelper.newline_character
            )
            + lrd_token.link_title_whitespace.count(ParserHelper.newline_character)
            + lrd_token.link_title_raw.count(ParserHelper.newline_character)
        )

    def premanage_container_tokens(self, token: MarkdownToken) -> None:
        if (
            self.container_token_stack
            and self.last_leaf_token
            and self.last_leaf_token.is_setext_heading
            and token.is_setext_heading_end
        ):
            self.bq_line_index[len(self.container_token_stack)] += 1

    def manage_container_tokens(self, token: MarkdownToken) -> None:
        """
        Manage the container tokens, especially the block quote indices.
        """
        if token.is_block_quote_start:
            self.container_token_stack.append(token)
            self.bq_line_index[len(self.container_token_stack)] = 0
        elif token.is_block_quote_end:
            del self.bq_line_index[len(self.container_token_stack)]
            del self.container_token_stack[-1]
        elif token.is_list_start:
            self.container_token_stack.append(token)
            self.bq_line_index[len(self.container_token_stack)] = 0
            self.list_adjust_map[len(self.container_token_stack)] = 1
        elif token.is_new_list_item:
            self.list_adjust_map[len(self.container_token_stack)] += 1
        elif token.is_list_end:
            del self.bq_line_index[len(self.container_token_stack)]
            del self.list_adjust_map[len(self.container_token_stack)]
            del self.container_token_stack[-1]
        elif self.container_token_stack:
            self.__manage_leaf_tokens(token)



"""
Module to provide for an encapsulation of the image element.
"""

import logging
from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.links.link_helper_properties import LinkHelperProperties
from pymarkdown.tokens.link_start_markdown_token import LinkStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.reference_markdown_token import ReferenceMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class ImageStartMarkdownToken(ReferenceMarkdownToken):
    """
    Class to provide for an encapsulation of the image element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        image_alt_text: str,
        text_from_blocks: str,
        line_number: int,
        column_number: int,
        lhp: LinkHelperProperties,
    ) -> None:
        self.__image_alt_text = image_alt_text
        ReferenceMarkdownToken.__init__(
            self,
            lhp,
            ImageStartMarkdownToken.get_markdown_token_type(),
            self.__image_alt_text,
            text_from_blocks,
            line_number=line_number,
            column_number=column_number,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_inline_image

    # pylint: enable=protected-access

    @property
    def image_alt_text(self) -> str:
        """
        Returns the text extracted from the blocks of the link, after processing.
        """
        return self.__image_alt_text

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            ImageStartMarkdownToken,
            ImageStartMarkdownToken.__rehydrate_inline_image,
            None,
        )

    @staticmethod
    def __rehydrate_inline_image(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the image text from the token.
        """
        _ = previous_token

        if context.block_stack[-1].is_inline_link:
            return ""
        inline_current_token = cast(ImageStartMarkdownToken, current_token)
        POGGER.debug(
            f">>>>>>>>:{ParserHelper.make_value_visible(inline_current_token)}:<<<<<"
        )
        rehydrated_text = (
            ImageStartMarkdownToken.rehydrate_inline_image_text_from_token(
                inline_current_token
            )
        )
        POGGER.debug(
            f">>>>>>>>:{ParserHelper.make_value_visible(rehydrated_text)}:<<<<<"
        )
        return LinkStartMarkdownToken.insert_leading_whitespace_at_newlines(
            context, rehydrated_text
        )

    @staticmethod
    def rehydrate_inline_image_text_from_token(
        image_token: "ImageStartMarkdownToken",
    ) -> str:
        """
        Given an image token, rehydrate it's original text from the token.
        """
        return f"!{LinkStartMarkdownToken.rehydrate_inline_link_text_from_token(image_token)}"

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            ImageStartMarkdownToken, ImageStartMarkdownToken.__handle_image_token, None
        )

    @classmethod
    def __handle_image_token(
        cls,
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state

        image_token = cast(ImageStartMarkdownToken, next_token)
        return "".join(
            [
                output_html,
                '<img src="',
                image_token.link_uri,
                '" alt="',
                image_token.image_alt_text,
                '" ',
                (
                    f'title="{image_token.link_title}" '
                    if image_token.link_title
                    else ""
                ),
                "/>",
            ]
        )



"""
Module to implement a plugin that looks for spaces within link labels.
"""

from typing import cast

from pymarkdown.general.constants import Constants
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.link_reference_definition_markdown_token import (
    LinkReferenceDefinitionMarkdownToken,
)
from pymarkdown.tokens.link_start_markdown_token import LinkStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd039(RulePlugin):
    """
    Class to implement a plugin that looks for spaces within link labels.
    """

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="no-space-in-links",
            plugin_id="MD039",
            plugin_enabled_by_default=True,
            plugin_description="Spaces inside link text",
            plugin_version="0.5.2",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md039.md",
            plugin_supports_fix=True,
        )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_inline_link or token.is_inline_image:
            link_token = cast(LinkStartMarkdownToken, token)
            stripped_text_from_blocks = link_token.text_from_blocks.strip(
                Constants.ascii_whitespace
            )
            if link_token.text_from_blocks != stripped_text_from_blocks:
                if context.in_fix_mode:
                    self.register_fix_token_request(
                        context,
                        token,
                        "next_token",
                        "text_from_blocks",
                        stripped_text_from_blocks,
                    )
                else:
                    self.report_next_token_error(context, token)
        elif token.is_link_reference_definition:
            link_def_token = cast(LinkReferenceDefinitionMarkdownToken, token)
            assert link_def_token.link_name_debug is not None
            stripped_text_from_blocks = link_def_token.link_name_debug.strip(
                Constants.ascii_whitespace
            )
            if link_def_token.link_name_debug != stripped_text_from_blocks:
                if context.in_fix_mode:
                    self.register_fix_token_request(
                        context,
                        token,
                        "next_token",
                        "link_name_debug",
                        stripped_text_from_blocks,
                    )
                else:
                    self.report_next_token_error(context, token)



"""
Module to provide for an encapsulation of the inline raw html element.
"""

import logging
from typing import Optional, Union, cast

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class RawHtmlMarkdownToken(InlineMarkdownToken):
    """
    Class to provide for an encapsulation of the inline raw html element.
    """

    def __init__(self, raw_tag: str, line_number: int, column_number: int) -> None:
        self.__raw_tag = raw_tag
        InlineMarkdownToken.__init__(
            self,
            MarkdownToken._token_inline_raw_html,
            self.__raw_tag,
            line_number=line_number,
            column_number=column_number,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_inline_raw_html

    # pylint: enable=protected-access

    @property
    def raw_tag(self) -> str:
        """
        Returns the text that is the raw html tag.
        """
        return self.__raw_tag

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            RawHtmlMarkdownToken,
            RawHtmlMarkdownToken.__rehydrate_inline_raw_html,
            None,
        )

    @staticmethod
    def __rehydrate_inline_raw_html(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the email raw html from the token.
        """
        _ = previous_token

        if context.block_stack[-1].is_inline_link:
            return ""

        current_raw_token = cast(RawHtmlMarkdownToken, current_token)
        raw_text = ParserHelper.remove_all_from_text(current_raw_token.raw_tag)

        if context.block_stack[-1].is_paragraph:
            block_paragraph_token = cast(
                ParagraphMarkdownToken, context.block_stack[-1]
            )
            POGGER.debug(
                f"raw_html>>before>>{ParserHelper.make_value_visible(raw_text)}"
            )
            block_paragraph_token.rehydrate_index += (
                ParserHelper.count_newlines_in_text(raw_text)
            )
            POGGER.debug(
                f"raw_html>>after>>{ParserHelper.make_value_visible(raw_text)}"
            )
        return f"<{raw_text}>"

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            RawHtmlMarkdownToken, RawHtmlMarkdownToken.__handle_raw_html_token, None
        )

    @staticmethod
    def __handle_raw_html_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state

        raw_html_token = cast(RawHtmlMarkdownToken, next_token)
        return "".join(
            [
                output_html,
                "<",
                ParserHelper.resolve_all_from_text(raw_html_token.raw_tag),
                ">",
            ]
        )

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "raw_tag" and isinstance(field_value, str):
            self.__raw_tag = field_value
            self._set_extra_data(field_value)
            return True
        return False



"""
Module to provide for an encapsulation of the link reference definition element.
"""

# pylint: disable=too-many-instance-attributes
from typing import List, Optional, Union, cast

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.links.link_reference_info import LinkReferenceInfo
from pymarkdown.links.link_reference_titles import LinkReferenceTitles
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class LinkReferenceDefinitionMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the link reference definition element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        did_add_definition: bool,
        extracted_whitespace: str,
        link_name: str,
        link_value: LinkReferenceTitles,
        link_debug: Optional[LinkReferenceInfo],
        position_marker: PositionMarker,
    ) -> None:
        self.__did_add_definition = did_add_definition
        self.__link_name = link_name

        if link_value:
            self.__link_destination, self.__link_title = (
                link_value.inline_link,
                link_value.inline_title,
            )
        else:
            self.__link_destination, self.__link_title = "", ""

        if link_debug:
            (
                self.__link_name_debug,
                self.__link_destination_whitespace,
                self.__link_destination_raw,
                self.__link_title_whitespace,
                self.__link_title_raw,
                self.__end_whitespace,
            ) = (
                (
                    ""
                    if link_debug.collected_destination == self.__link_name
                    else link_debug.collected_destination
                ),
                link_debug.line_destination_whitespace,
                (
                    ""
                    if link_debug.inline_raw_link == self.__link_destination
                    else link_debug.inline_raw_link
                ),
                link_debug.line_title_whitespace,
                (
                    ""
                    if link_debug.inline_raw_title == self.__link_title
                    else link_debug.inline_raw_title
                ),
                link_debug.end_whitespace,
            )
        else:
            self.__link_name_debug = self.__link_destination_whitespace = (
                self.__link_destination_raw
            ) = self.__link_title_whitespace = self.__link_title_raw = (
                self.__end_whitespace
            ) = ""

        extra_data = self.__validate_proper_fields_are_valid(extracted_whitespace)
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_link_reference_definition,
            extra_data,
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_link_reference_definition

    # pylint: enable=protected-access

    def __validate_proper_fields_are_valid(self, extracted_whitespace: str) -> str:
        assert self.__end_whitespace is not None, "This field should be defined."
        assert self.__link_title_raw is not None, "This field should be defined."
        assert self.__link_title is not None, "This field should be defined."
        assert self.__link_title_whitespace is not None, "This field should be defined."
        assert self.__link_destination_raw is not None, "This field should be defined."
        assert self.__link_destination is not None, "This field should be defined."
        assert (
            self.__link_destination_whitespace is not None
        ), "This field should be defined."
        assert self.__link_name_debug is not None, "This field should be defined."
        return MarkdownToken.extra_data_separator.join(
            [
                str(self.did_add_definition),
                extracted_whitespace,
                self.__link_name,
                self.__link_name_debug,
                self.__link_destination_whitespace,
                self.__link_destination,
                self.__link_destination_raw,
                self.__link_title_whitespace,
                self.__link_title,
                self.__link_title_raw,
                self.__end_whitespace,
            ]
        )

    # pylint: enable=too-many-arguments
    @property
    def did_add_definition(self) -> bool:
        """
        Returns an indication of whether the definition was actually added.
        """
        return self.__did_add_definition

    @property
    def end_whitespace(self) -> Optional[str]:
        """
        Returns any whitespace that was extracted after the processing of this element occurred.
        """
        return self.__end_whitespace

    @property
    def link_name(self) -> str:
        """
        Returns the name of the link that was defined.
        """
        return self.__link_name

    @property
    def link_name_debug(self) -> Optional[str]:
        """
        Returns the name of the link that was defined, in debug form.
        """
        return self.__link_name_debug

    @property
    def link_destination_whitespace(self) -> Optional[str]:
        """
        Returns the whitespace that occurs before the link destination.
        """
        return self.__link_destination_whitespace

    @property
    def link_destination(self) -> Optional[str]:
        """
        Returns the destination (URI) of the link that was defined.
        """
        return self.__link_destination

    @property
    def link_destination_raw(self) -> Optional[str]:
        """
        Returns the destination (URI) of the link that was defined, in raw form.
        """
        return self.__link_destination_raw

    @property
    def link_title(self) -> Optional[str]:
        """
        Returns the title of the link that was defined.
        """
        return self.__link_title

    @property
    def link_title_raw(self) -> Optional[str]:
        """
        Returns the title of the link that was defined, in raw form.
        """
        return self.__link_title_raw

    @property
    def link_title_whitespace(self) -> Optional[str]:
        """
        Returns the whitespace that occurs after the link title.
        """
        return self.__link_title_whitespace

    # pylint: disable=too-many-return-statements
    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "link_destination_whitespace" and isinstance(field_value, str):
            self.__link_destination_whitespace = field_value
            extra_data = self.__validate_proper_fields_are_valid(
                self.extracted_whitespace
            )
            super()._set_extra_data(extra_data)
            return True
        if field_name == "link_title_whitespace" and isinstance(field_value, str):
            self.__link_title_whitespace = field_value
            extra_data = self.__validate_proper_fields_are_valid(
                self.extracted_whitespace
            )
            super()._set_extra_data(extra_data)
            return True
        if field_name == "link_name" and isinstance(field_value, str):
            self.__link_name = field_value
            extra_data = self.__validate_proper_fields_are_valid(
                self.extracted_whitespace
            )
            super()._set_extra_data(extra_data)
            return True
        if field_name == "link_name_debug" and isinstance(field_value, str):
            self.__link_name_debug = field_value
            extra_data = self.__validate_proper_fields_are_valid(
                self.extracted_whitespace
            )
            super()._set_extra_data(extra_data)
            return True
        if field_name == "link_title" and isinstance(field_value, str):
            self.__link_title = field_value
            extra_data = self.__validate_proper_fields_are_valid(
                self.extracted_whitespace
            )
            super()._set_extra_data(extra_data)
            return True
        if field_name == "link_title_raw" and isinstance(field_value, str):
            self.__link_title_raw = field_value
            extra_data = self.__validate_proper_fields_are_valid(
                self.extracted_whitespace
            )
            super()._set_extra_data(extra_data)
            return True
        return super()._modify_token(field_name, field_value)

    # pylint: enable=too-many-return-statements

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            LinkReferenceDefinitionMarkdownToken,
            LinkReferenceDefinitionMarkdownToken.__rehydrate_link_reference_definition,
            None,
        )

    @staticmethod
    def __rehydrate_link_reference_definition(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, context)

        current_lrd_token = cast(LinkReferenceDefinitionMarkdownToken, current_token)
        link_title_text = (
            current_lrd_token.link_title_raw or current_lrd_token.link_title
        )
        link_destination_text = (
            current_lrd_token.link_destination_raw or current_lrd_token.link_destination
        )
        assert (
            current_lrd_token.link_destination_whitespace is not None
        ), "This field should be defined."
        assert link_destination_text is not None, "This field should be defined."
        assert (
            current_lrd_token.link_title_whitespace is not None
        ), "This field should be defined."
        assert link_title_text is not None, "This field should be defined."
        assert (
            current_lrd_token.end_whitespace is not None
        ), "This field should be defined."
        link_def_parts: List[str] = [
            current_lrd_token.extracted_whitespace,
            "[",
            current_lrd_token.link_name_debug or current_lrd_token.link_name,
            "]:",
            current_lrd_token.link_destination_whitespace,
            link_destination_text,
            current_lrd_token.link_title_whitespace,
            link_title_text,
            current_lrd_token.end_whitespace,
            ParserHelper.newline_character,
        ]

        return "".join(link_def_parts)

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            LinkReferenceDefinitionMarkdownToken,
            LinkReferenceDefinitionMarkdownToken.__handle_link_reference_definition_token,
            None,
        )

    @staticmethod
    def __handle_link_reference_definition_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return output_html


# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that looks for multiple spaces after the hash
mark on an atx heading.
"""

from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd019(RulePlugin):
    """
    Class to implement a plugin that looks for multiple spaces after the hash
    mark on an atx heading.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__atx_heading_token: Optional[AtxHeadingMarkdownToken] = None

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="no-multiple-space-atx",
            plugin_id="MD019",
            plugin_enabled_by_default=True,
            plugin_description="Multiple spaces are present after hash character on Atx Heading.",
            plugin_version="0.5.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md019.md",
            plugin_supports_fix=True,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__atx_heading_token = None

    def __report(
        self, context: PluginScanContext, text_token: TextMarkdownToken
    ) -> None:
        assert self.__atx_heading_token is not None
        if context.in_fix_mode:
            self.register_fix_token_request(
                context,
                text_token,
                "next_token",
                "extracted_whitespace",
                " ",
            )
        else:
            self.report_next_token_error(context, self.__atx_heading_token)

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_atx_heading:
            atx_token = cast(AtxHeadingMarkdownToken, token)
            if not atx_token.remove_trailing_count:
                self.__atx_heading_token = atx_token
        elif token.is_paragraph_end:
            self.__atx_heading_token = None
        elif token.is_text:
            text_token = cast(TextMarkdownToken, token)
            if self.__atx_heading_token:
                resolved_extracted_whitespace = ParserHelper.remove_all_from_text(
                    text_token.extracted_whitespace
                )
                if "\t" in resolved_extracted_whitespace:
                    start_index = (
                        self.__atx_heading_token.column_number
                        - 1
                        + self.__atx_heading_token.hash_count
                    )
                    resolved_extracted_whitespace = TabHelper.detabify_string(
                        resolved_extracted_whitespace, start_index
                    )
                if len(resolved_extracted_whitespace) > 1:
                    self.__report(context, text_token)
            self.__atx_heading_token = None



"""
Module to provide a high level abstraction of the extensions.
"""

from typing import TYPE_CHECKING, Dict, Optional, cast

from pymarkdown.extensions.disallowed_raw_html import MarkdownDisallowRawHtmlExtension

if TYPE_CHECKING:  # pragma: no cover
    from pymarkdown.extension_manager.extension_manager import ExtensionManager


class ParseBlockPassProperties:
    """
    Class to provide a high level abstraction of the extensions.
    """

    def __init__(self, extension_manager: "ExtensionManager") -> None:
        """
        Note: is_strike_through_enabled is not included in this list as it exclusively
              effects the inline processing which does not require these properties
        """

        (
            self.__front_matter_enabled,
            self.__pragmas_enabled,
            self.__disallow_raw_html_enabled,
            self.__task_lists_enabled,
            self.__tables_enabled,
        ) = (
            extension_manager.is_front_matter_enabled,
            extension_manager.is_linter_pragmas_enabled,
            extension_manager.is_disallow_raw_html_enabled,
            extension_manager.is_task_list_items_enabled,
            extension_manager.is_tables_enabled,
        )
        self.pragma_lines: Dict[int, str] = {}

        self.__disable_raw_html_extension = None
        if self.__disallow_raw_html_enabled:
            ext_instance = extension_manager.get_extension_instance(
                MarkdownDisallowRawHtmlExtension().get_identifier()
            )
            self.__disable_raw_html_extension = cast(
                MarkdownDisallowRawHtmlExtension, ext_instance
            )

    @property
    def is_front_matter_enabled(self) -> bool:
        """
        Returns whether front matter parsing is enabled.
        """
        return self.__front_matter_enabled

    @property
    def is_pragmas_enabled(self) -> bool:
        """
        Returns whether pragma parsing is enabled.
        """
        return self.__pragmas_enabled

    @property
    def is_disallow_raw_html_enabled(self) -> bool:
        """
        Returns whether disallow raw html is enabled.
        """
        return self.__disallow_raw_html_enabled

    @property
    def disallow_raw_html(self) -> Optional[MarkdownDisallowRawHtmlExtension]:
        return self.__disable_raw_html_extension

    @property
    def is_task_lists_enabled(self) -> bool:
        """
        Returns whether task lists are enabled.
        """
        return self.__task_lists_enabled

    @property
    def is_tables_enabled(self) -> bool:
        """
        Returns whether tables are enabled.
        """
        return self.__tables_enabled



"""
Module to implement a plugin that
"""

from typing import List, Set, Tuple, cast

from pymarkdown.extensions.disallowed_raw_html import MarkdownDisallowRawHtmlExtension
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.utils.container_token_manager import ContainerTokenManager
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RulePml100(RulePlugin):
    """
    Class to implement a plugin that
    """

    def __init__(self) -> None:
        super().__init__()
        self.__container_manager = ContainerTokenManager()
        self.__in_block = False
        self.__disallowed_tag_names: Set[str] = set()
        self.__modify_tag_names: str = ""

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="disallowed-html",
            plugin_id="PML100",
            plugin_enabled_by_default=False,
            plugin_description="Disallowed HTML",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_pml100.md",
            plugin_configuration=None,
            plugin_supports_fix=False,
        )

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__disallowed_tag_names = {
            "title",
            "textarea",
            "style",
            "xmp",
            "iframe",
            "noembed",
            "noframes",
            "script",
            "plaintext",
        }
        self.__modify_tag_names = (
            self.plugin_configuration.get_string_property_with_default(
                "change_tag_names", ""
            )
        )
        if self.__modify_tag_names:
            tag_config_name = "plugins.disallowed-html.change_tag_names"
            for next_tag_part in self.__modify_tag_names.split(","):
                next_tag_part = next_tag_part.strip(" ")
                if not next_tag_part:
                    raise ValueError(
                        f"Configuration item '{tag_config_name}' contains at least one empty string."
                    )
                if next_tag_part[0] not in ("+", "-"):
                    raise ValueError(
                        f"Configuration item '{tag_config_name}' elements must either start with '+' or '-'."
                    )
                remaining_tag_part = next_tag_part[1:]
                if not MarkdownDisallowRawHtmlExtension.is_valid_tag_name(
                    remaining_tag_part
                ):
                    raise ValueError(
                        f"Configuration item '{tag_config_name}' contains an element '{remaining_tag_part}' that is not a valid tag name."
                    )
                if next_tag_part[0] == "+":
                    self.__disallowed_tag_names.add(remaining_tag_part.lower())
                else:
                    self.__disallowed_tag_names.remove(remaining_tag_part.lower())

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("change_tag_names", self.__modify_tag_names),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__in_block = False
        self.__container_manager.clear()

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_html_block:
            self.__in_block = True
        elif token.is_html_block_end:
            self.__in_block = False
        elif self.__in_block and token.is_text:
            self.__scan_text_block(context, cast(TextMarkdownToken, token))
        elif token.is_inline_raw_html:
            self.__scan_text_token(context, cast(RawHtmlMarkdownToken, token))

        self.__container_manager.manage_container_tokens(token)

    def __find_delta_column(self, delta_text: str) -> Tuple[int, int]:
        if "\n" in delta_text:
            split_delta_text = delta_text.split(ParserHelper.newline_character)
            column_number_delta = -(len(split_delta_text[-1]) + 1)
            return len(split_delta_text) - 1, column_number_delta
        return 0, len(delta_text)

    def __scan_text_block(
        self, context: PluginScanContext, text_token: TextMarkdownToken
    ) -> None:
        ind = text_token.token_text.find("<")
        while ind >= 0:
            (
                after_text_index,
                collected_text,
            ) = ParserHelper.collect_until_one_of_characters(
                text_token.token_text, ind + 1, " /<>"
            )
            if collected_text in self.__disallowed_tag_names:
                extra_data = f"Tag Name: {collected_text}"
                text_so_far = text_token.token_text[:ind]
                line_number_delta, column_number_delta = self.__find_delta_column(
                    text_so_far
                )

                if len(self.__container_manager.container_token_stack) > 0:
                    if container_base_column := self.__calculate_base_column(
                        column_number_delta, text_so_far
                    ):
                        column_number_delta -= container_base_column

                self.report_next_token_error(
                    context,
                    text_token,
                    extra_error_information=extra_data,
                    line_number_delta=line_number_delta,
                    column_number_delta=column_number_delta,
                )
            ind = text_token.token_text.find("<", after_text_index)

    def __scan_text_token(
        self, context: PluginScanContext, text_token: RawHtmlMarkdownToken
    ) -> None:
        assert text_token.extra_data is not None
        split_text = text_token.extra_data.split(" ", 1)
        collected_text = split_text[0]
        if collected_text.endswith("/"):
            collected_text = collected_text[:-1]
        if collected_text in self.__disallowed_tag_names:
            extra_data = f"Tag Name: {collected_text}"
            self.report_next_token_error(
                context, text_token, extra_error_information=extra_data
            )

    def __calculate_base_column(
        self, column_number_delta: int, text_so_far: str
    ) -> int:
        container_base_column = 0
        stack_index = len(self.__container_manager.container_token_stack) - 1
        top_one = self.__container_manager.container_token_stack[-1]
        if top_one.is_list_start:
            if column_number_delta < 0:
                list_token = cast(
                    ListStartMarkdownToken,
                    self.__container_manager.container_token_stack[stack_index],
                )
                container_base_column = list_token.indent_level
        else:
            assert top_one.is_block_quote_start
            container_base_column = 0
            count_of_block_quote_characters = text_so_far.count(">")
            if count_of_block_quote_characters > 0:
                bq_index = (
                    self.__container_manager.bq_line_index[stack_index + 1]
                    + count_of_block_quote_characters
                )
                block_quote_token = cast(
                    BlockQuoteMarkdownToken,
                    self.__container_manager.container_token_stack[stack_index],
                )
                assert block_quote_token.bleading_spaces is not None
                split_leading_spaces = block_quote_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
                container_base_column = len(split_leading_spaces[bq_index])
                assert column_number_delta < 0
                # if column_number_delta >= 0:
                #     container_base_column = -container_base_column

        return container_base_column



"""
Module to calculate whether to close a list when starting a new list.
"""

import logging
from typing import List, Tuple, cast

from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken, StackToken

POGGER = ParserLogger(logging.getLogger(__name__))


class ListBlockCanCloseHelper:
    """
    Class to calculate whether to close a list when starting a new list.
    """

    @staticmethod
    def calculate_can_remove_list(
        parser_state: ParserState, current_start_index: int
    ) -> bool:
        POGGER.debug_with_visible_whitespace(
            "parser_state.token_stack>$", parser_state.token_stack
        )
        POGGER.debug_with_visible_whitespace(
            "current_start_index>$", current_start_index
        )
        if len(parser_state.token_stack) <= 2:
            return False

        stack_index = len(parser_state.token_stack) - 1
        while stack_index and not parser_state.token_stack[stack_index].is_list:
            stack_index -= 1
        assert stack_index, "Stack index must be positive. (Document = index 0)"
        last_stack_index = stack_index
        assert parser_state.token_stack[
            stack_index
        ].is_list, "Current block must be a list."
        stack_index -= 1
        while stack_index and not parser_state.token_stack[stack_index].is_list:
            stack_index -= 1
        if not stack_index:
            return False

        stack_index_token = cast(ListStackToken, parser_state.token_stack[stack_index])
        last_stack_index_token = cast(
            ListStackToken, parser_state.token_stack[last_stack_index]
        )
        POGGER.debug("parser_state.token_stack[i]>$", stack_index_token)
        previous_list_start_index = stack_index_token.indent_level
        previous_list_end_index = (
            previous_list_start_index
            + len(stack_index_token.list_character)
            + stack_index_token.ws_after_marker
        )
        POGGER.debug(
            "token_stack[i]:  previous_list_start_index=$,previous_list_end_index=$",
            previous_list_start_index,
            previous_list_end_index,
        )
        with_previous_list_bounds = (
            previous_list_start_index <= current_start_index < previous_list_end_index
        )

        adjusted_index = (
            last_stack_index_token.indent_level
            - last_stack_index_token.ws_before_marker
            if last_stack_index_token.is_ordered_list
            else last_stack_index_token.indent_level - 2
        )

        POGGER.debug(
            "adjusted_index>$ > parser_state.token_stack[stack_index].indent_level>$ ) "
            + "or not with_previous_list_bounds=$",
            adjusted_index,
            stack_index_token.indent_level,
            with_previous_list_bounds,
        )
        return (
            adjusted_index > stack_index_token.indent_level
            or not with_previous_list_bounds
        )

    @staticmethod
    def close_required_lists(
        parser_state: ParserState,
        allow_list_removal: bool,
        balancing_tokens: List[MarkdownToken],
        new_stack: StackToken,
    ) -> None:
        assert (
            new_stack.matching_markdown_token is not None
        ), "New stack token must have a matching markdown token."
        matching_column_number = new_stack.matching_markdown_token.column_number

        (
            list_count,
            parent_indent_level,
        ) = ListBlockCanCloseHelper.__close_required_lists_calc(parser_state)

        POGGER.debug(
            "matching_column_number=$ <= parent_indent_level=$ and allow_list_removal=$ and list_count=$ > 1",
            matching_column_number,
            parent_indent_level,
            allow_list_removal,
            list_count,
        )
        while (
            matching_column_number <= parent_indent_level
            and allow_list_removal
            and list_count > 1
        ):
            stack_index = len(parser_state.token_stack) - 2
            while stack_index and not parser_state.token_stack[stack_index].is_list:
                stack_index -= 1
            last_stack_index = parser_state.token_stack.index(
                parser_state.token_stack[-1]
            )
            POGGER.debug("parser_state.token_stack>>$", parser_state.token_stack)
            POGGER.debug("last_stack_index>>$", last_stack_index)
            POGGER.debug("stack_index>>$", stack_index)
            last_stack_index = stack_index + 1

            close_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=last_stack_index,
                include_lists=True,
                include_block_quotes=True,
            )
            assert close_tokens, "At least one token must have been returned."
            balancing_tokens.extend(close_tokens)
            POGGER.debug("close_tokens>>$", close_tokens)
            POGGER.debug("parser_state.token_stack>>$", parser_state.token_stack)
            assert parser_state.token_stack[-1].is_list, "Current block must be a list."

            (
                list_count,
                parent_indent_level,
            ) = ListBlockCanCloseHelper.__close_required_lists_calc(parser_state)
            POGGER.debug(
                "matching_column_number=$ < parent_indent_level=$ and "
                + "allow_list_removal=$ and list_count=$ > 1",
                matching_column_number,
                parent_indent_level,
                allow_list_removal,
                list_count,
            )

    @staticmethod
    def __close_required_lists_calc(parser_state: ParserState) -> Tuple[int, int]:
        token_stack_index = len(parser_state.token_stack) - 1
        found_list_tokens = []
        while token_stack_index:
            if parser_state.token_stack[token_stack_index].is_list:
                found_list_tokens.append(parser_state.token_stack[token_stack_index])
            token_stack_index -= 1

        parent_indent_level = -1
        if len(found_list_tokens) > 1:
            list_token = cast(
                ListStartMarkdownToken, found_list_tokens[1].matching_markdown_token
            )
            parent_indent_level = list_token.indent_level
        return len(found_list_tokens), parent_indent_level



"""
Class to provide a "grab bag" for commonly used properties for the Container Block Processor.
"""

import logging
from typing import Any, List, Optional, Tuple

from typing_extensions import Protocol

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_indices import ContainerIndices
from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import StackToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class ParseForContainerBlocksProtocol(Protocol):
    """
    Protocol to do further parsing for container blocks.
    """

    # pylint: disable=too-many-arguments
    def __call__(  # noqa: E704
        self,
        parser_state: ParserState,
        position_marker: PositionMarker,
        ignore_link_definition_start: bool,
        ignore_table_start: bool,
        parser_properties: ParseBlockPassProperties,
        container_start_bq_count: int,
        container_depth: int,
        adjusted_block_index: Optional[int],
        initial_block_quote_count: Optional[int],
        original_line: Optional[str],
    ) -> Tuple[
        List[MarkdownToken],
        Optional[str],
        Optional[int],
        Optional[RequeueLineInfo],
        bool,
        bool,
    ]: ...  # pragma: no cover

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
class ContainerGrabBag:
    """
    Class to provide for a grab bag of values instead of passing
    them around one by one.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        parser_state: ParserState,
        container_depth: int,
        initial_block_quote_count: Optional[int],
        adjusted_block_index: Optional[int],
        container_start_bq_count: int,
        parser_properties: ParseBlockPassProperties,
        ignore_link_definition_start: bool,
        ignore_table_start: bool,
        original_line: str,
        recurse_fn: ParseForContainerBlocksProtocol,
    ) -> None:
        # Functions
        self.__recurse_fn = recurse_fn

        # Booleans
        self.__did_blank: bool = False
        self.__was_paragraph_continuation: bool = False
        self.__do_skip_containers_before_leaf_blocks: bool = False
        self.__was_indent_already_processed: bool = False
        self.__do_force_leaf_token_parse: bool = False
        self.__did_indent_processing: bool = False
        self.__have_pre_processed_indent: bool = False
        self.__can_continue: bool = False
        self.__do_force_list_continuation = False

        # Integers
        self.__indent_used_by_container: int = -1
        self.__indent_already_processed: int = -1
        self.__last_block_quote_index: int = -1
        self.__last_list_start_index: int = -1
        self.__start_index: int = -1

        # Optional integers
        self.__removed_chars_at_start_of_line: Optional[int] = None

        # Strings
        self.__adj_line_to_parse: str = ""
        self.__line_to_parse: str = ""

        # Optional strings
        self.__weird_adjusted_text: Optional[str] = None
        self.__indent_used_by_list: Optional[str] = None
        self.__text_removed_by_container: Optional[str] = None
        self.__extracted_whitespace: Optional[str] = None
        self.__adj_ws: Optional[str] = None
        self.__bogus: Optional[str] = None

        # Others
        self.__block_quote_data = BlockQuoteData(-1, -1)
        self.__end_container_indices = ContainerIndices(-1, -1, -1)
        self.__requeue_line_info: Optional[RequeueLineInfo] = None
        self.__current_container_blocks: List[StackToken] = []
        self.__leaf_tokens: List[MarkdownToken] = []
        self.__container_level_tokens: List[MarkdownToken] = []

        # Read only values
        self.__container_depth = container_depth
        self.__initial_block_quote_count = initial_block_quote_count
        self.__adjusted_block_index = adjusted_block_index
        self.__container_start_bq_count = container_start_bq_count
        self.__parser_properties = parser_properties
        self.__do_ignore_link_definition_start = ignore_link_definition_start
        self.__do_ignore_table_start = ignore_table_start
        self.__original_line = original_line
        self.__is_para_continue = (
            bool(
                parser_state
                and parser_state.token_stack
                and len(parser_state.token_stack) >= 2
                and parser_state.token_stack[-1].is_paragraph
            )
            and not parser_state.token_document[-1].is_blank_line
        )
        self.__log_initial_values()

    def __log_initial_values(self) -> None:
        # Booleans
        self.__log_read_write_value("did_blank", self.__did_blank)
        self.__log_read_write_value(
            "was_paragraph_continuation", self.__was_paragraph_continuation
        )
        self.__log_read_write_value(
            "do_skip_containers_before_leaf_blocks",
            self.__do_skip_containers_before_leaf_blocks,
        )
        self.__log_read_write_value(
            "was_indent_already_processed", self.__was_indent_already_processed
        )
        self.__log_read_write_value(
            "do_force_leaf_token_parse", self.__do_force_leaf_token_parse
        )
        self.__log_read_write_value(
            "did_indent_processing", self.__did_indent_processing
        )
        self.__log_read_write_value(
            "have_pre_processed_indent", self.__have_pre_processed_indent
        )
        self.__log_read_write_value("can_continue", self.__can_continue)
        self.__log_read_write_value(
            "do_force_list_continuation", self.__do_force_list_continuation
        )

        # Integers
        self.__log_read_write_value(
            "indent_used_by_container", self.__indent_used_by_container
        )
        self.__log_read_write_value(
            "indent_already_processed", self.__indent_already_processed
        )
        self.__log_read_write_value(
            "last_block_quote_index", self.__last_block_quote_index
        )
        self.__log_read_write_value(
            "last_list_start_index", self.__last_list_start_index
        )
        self.__log_read_write_value("start_index", self.__start_index)

        # Optional integers
        self.__log_read_write_value(
            "removed_chars_at_start_of_line", self.__removed_chars_at_start_of_line
        )

        # Strings
        self.__log_read_write_value("adj_line_to_parse", self.__adj_line_to_parse)
        self.__log_read_write_value("line_to_parse", self.__line_to_parse)

        # Optional strings
        self.__log_read_write_value("weird_adjusted_text", self.__weird_adjusted_text)
        self.__log_read_write_value("indent_used_by_list", self.__indent_used_by_list)
        self.__log_read_write_value(
            "text_removed_by_container", self.__text_removed_by_container
        )
        self.__log_read_write_value("extracted_whitespace", self.__extracted_whitespace)
        self.__log_read_write_value("adj_ws", self.__adj_ws)
        self.__log_read_write_value("bogus", self.__bogus)

        # Others
        self.__log_read_write_value("block_quote_data", self.__block_quote_data)
        self.__log_read_write_value(
            "end_container_indices", self.__end_container_indices
        )
        self.__log_read_write_value("requeue_line_info", self.__requeue_line_info)
        self.__log_read_write_value("leaf_tokens", self.__leaf_tokens)
        self.__log_read_write_value(
            "container_level_tokens", self.__container_level_tokens
        )

        # Read only values
        self.__log_read_only_value("container_depth", self.__container_depth)
        self.__log_read_only_value(
            "initial_block_quote_count", self.__initial_block_quote_count
        )
        self.__log_read_only_value("adjusted_block_index", self.__adjusted_block_index)
        self.__log_read_only_value(
            "container_start_bq_count", self.__container_start_bq_count
        )
        self.__log_read_only_value("parser_properties", self.__parser_properties)
        self.__log_read_only_value(
            "do_ignore_link_definition_start", self.__do_ignore_link_definition_start
        )
        self.__log_read_only_value(
            "do_ignore_table_start", self.__do_ignore_table_start
        )
        self.__log_read_only_value("original_line", self.__original_line)
        self.__log_read_only_value("is_para_continue", self.__is_para_continue)

    @staticmethod
    def __log_read_only_value(value_name: str, value_object: Any) -> None:
        POGGER.debug("CGB(ReadOnly): '$'=:$:", value_name, value_object)

    @staticmethod
    def __log_read_write_value(value_name: str, value_object: Any) -> None:
        POGGER.debug("CGB(ReadWrite): '$'=:$:", value_name, value_object)

    # pylint: enable=too-many-arguments
    @property
    def is_para_continue(self) -> bool:
        """
        Xxx
        """
        return self.__is_para_continue

    @property
    def original_line(self) -> str:
        """
        Get the original line that was passed to be processed.
        """
        return self.__original_line

    @property
    def container_depth(self) -> int:
        """
        Get the container depth of this pass through the line.
        """
        return self.__container_depth

    @property
    def initial_block_quote_count(self) -> Optional[int]:
        """
        Get any initial block quote informatin.
        """
        return self.__initial_block_quote_count

    @property
    def adjusted_block_index(self) -> Optional[int]:
        """
        Index of the block start.
        """
        return self.__adjusted_block_index

    @property
    def container_start_bq_count(self) -> Optional[int]:
        """
        Xxx
        """
        return self.__container_start_bq_count

    @property
    def parser_properties(self) -> ParseBlockPassProperties:
        """
        Xxx
        """
        return self.__parser_properties

    @property
    def do_ignore_link_definition_start(self) -> bool:
        """
        Xxx
        """
        return self.__do_ignore_link_definition_start

    @property
    def do_ignore_table_start(self) -> bool:
        """
        Xxx
        """
        return self.__do_ignore_table_start

    @property
    def do_force_list_continuation(self) -> bool:
        """
        Xxx
        """
        return self.__do_force_list_continuation

    @do_force_list_continuation.setter
    def do_force_list_continuation(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__do_force_list_continuation:
            self.__log_read_write_value("do_force_list_continuation", value)
        self.__do_force_list_continuation = value

    @property
    def did_blank(self) -> bool:
        """
        Xxx
        """
        return self.__did_blank

    @did_blank.setter
    def did_blank(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__did_blank:
            self.__log_read_write_value("did_blank", value)
            self.__did_blank = value

    @property
    def was_paragraph_continuation(self) -> bool:
        """
        Xxx
        """
        return self.__was_paragraph_continuation

    @was_paragraph_continuation.setter
    def was_paragraph_continuation(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__was_paragraph_continuation:
            self.__log_read_write_value("was_paragraph_continuation", value)
            self.__was_paragraph_continuation = value

    @property
    def do_skip_containers_before_leaf_blocks(self) -> bool:
        """
        Xxx
        """
        return self.__do_skip_containers_before_leaf_blocks

    @do_skip_containers_before_leaf_blocks.setter
    def do_skip_containers_before_leaf_blocks(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__do_skip_containers_before_leaf_blocks:
            self.__log_read_write_value("do_skip_containers_before_leaf_blocks", value)
            self.__do_skip_containers_before_leaf_blocks = value

    @property
    def was_indent_already_processed(self) -> bool:
        """
        Xxx
        """
        return self.__was_indent_already_processed

    @was_indent_already_processed.setter
    def was_indent_already_processed(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__was_indent_already_processed:
            self.__log_read_write_value("was_indent_already_processed", value)
            self.__was_indent_already_processed = value

    @property
    def do_force_leaf_token_parse(self) -> bool:
        """
        Xxx
        """
        return self.__do_force_leaf_token_parse

    @do_force_leaf_token_parse.setter
    def do_force_leaf_token_parse(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__do_force_leaf_token_parse:
            self.__log_read_write_value("do_force_leaf_token_parse", value)
            self.__do_force_leaf_token_parse = value

    @property
    def did_indent_processing(self) -> bool:
        """
        Xxx
        """
        return self.__did_indent_processing

    @did_indent_processing.setter
    def did_indent_processing(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__did_indent_processing:
            self.__log_read_write_value("did_indent_processing", value)
            self.__did_indent_processing = value

    @property
    def have_pre_processed_indent(self) -> bool:
        """
        Xxx
        """
        return self.__have_pre_processed_indent

    @have_pre_processed_indent.setter
    def have_pre_processed_indent(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__have_pre_processed_indent:
            self.__log_read_write_value("have_pre_processed_indent", value)
            self.__have_pre_processed_indent = value

    @property
    def can_continue(self) -> bool:
        """
        Xxx
        """
        return self.__can_continue

    @can_continue.setter
    def can_continue(self, value: bool) -> None:
        """
        xxx
        """
        if value != self.__can_continue:
            self.__log_read_write_value("can_continue", value)
            self.__can_continue = value

    @property
    def indent_used_by_container(self) -> int:
        """
        Xxx
        """
        return self.__indent_used_by_container

    @indent_used_by_container.setter
    def indent_used_by_container(self, value: int) -> None:
        """
        xxx
        """
        if value != self.__indent_used_by_container:
            self.__log_read_write_value("indent_used_by_container", value)
            self.__indent_used_by_container = value

    @property
    def indent_already_processed(self) -> int:
        """
        Xxx
        """
        return self.__indent_already_processed

    @indent_already_processed.setter
    def indent_already_processed(self, value: int) -> None:
        """
        xxx
        """
        if value != self.__indent_already_processed:
            self.__log_read_write_value("indent_already_processed", value)
            self.__indent_already_processed = value

    @property
    def last_block_quote_index(self) -> int:
        """
        Xxx
        """
        return self.__last_block_quote_index

    @last_block_quote_index.setter
    def last_block_quote_index(self, value: int) -> None:
        """
        xxx
        """
        if value != self.__last_block_quote_index:
            self.__log_read_write_value("last_block_quote_index", value)
            self.__last_block_quote_index = value

    @property
    def last_list_start_index(self) -> int:
        """
        Xxx
        """
        return self.__last_list_start_index

    @last_list_start_index.setter
    def last_list_start_index(self, value: int) -> None:
        """
        xxx
        """
        if value != self.__last_list_start_index:
            self.__log_read_write_value("last_list_start_index", value)
            self.__last_list_start_index = value

    @property
    def start_index(self) -> int:
        """
        Xxx
        """
        return self.__start_index

    @start_index.setter
    def start_index(self, value: int) -> None:
        """
        xxx
        """
        if value != self.__start_index:
            self.__log_read_write_value("start_index", value)
            self.__start_index = value

    @property
    def line_to_parse(self) -> str:
        """
        Xxx
        """
        return self.__line_to_parse

    @line_to_parse.setter
    def line_to_parse(self, value: str) -> None:
        """
        xxx
        """
        if value != self.__line_to_parse:
            self.__log_read_write_value("line_to_parse", value)
            self.__line_to_parse = value

    @property
    def adj_line_to_parse(self) -> str:
        """
        Xxx
        """
        return self.__adj_line_to_parse

    @adj_line_to_parse.setter
    def adj_line_to_parse(self, value: str) -> None:
        """
        xxx
        """
        if value != self.__adj_line_to_parse:
            self.__log_read_write_value("adj_line_to_parse", value)
            self.__adj_line_to_parse = value

    @property
    def weird_adjusted_text(self) -> Optional[str]:
        """
        Xxx
        """
        return self.__weird_adjusted_text

    @weird_adjusted_text.setter
    def weird_adjusted_text(self, value: Optional[str]) -> None:
        """
        xxx
        """
        if value != self.__weird_adjusted_text:
            self.__log_read_write_value("weird_adjusted_text", value)
            self.__weird_adjusted_text = value

    @property
    def indent_used_by_list(self) -> Optional[str]:
        """
        Xxx
        """
        return self.__indent_used_by_list

    @indent_used_by_list.setter
    def indent_used_by_list(self, value: Optional[str]) -> None:
        """
        xxx
        """
        if value != self.__indent_used_by_list:
            self.__log_read_write_value("indent_used_by_list", value)
            self.__indent_used_by_list = value

    @property
    def text_removed_by_container(self) -> Optional[str]:
        """
        Xxx
        """
        return self.__text_removed_by_container

    @text_removed_by_container.setter
    def text_removed_by_container(self, value: Optional[str]) -> None:
        """
        xxx
        """
        if value != self.__text_removed_by_container:
            self.__log_read_write_value("text_removed_by_container", value)
            self.__text_removed_by_container = value

    @property
    def extracted_whitespace(self) -> Optional[str]:
        """
        Xxx
        """
        return self.__extracted_whitespace

    @extracted_whitespace.setter
    def extracted_whitespace(self, value: Optional[str]) -> None:
        """
        xxx
        """
        if value != self.__extracted_whitespace:
            self.__log_read_write_value("extracted_whitespace", value)
            self.__extracted_whitespace = value

    @property
    def adj_ws(self) -> Optional[str]:
        """
        Xxx
        """
        return self.__adj_ws

    @adj_ws.setter
    def adj_ws(self, value: Optional[str]) -> None:
        """
        xxx
        """
        if value != self.__adj_ws:
            self.__log_read_write_value("adj_ws", value)
            self.__adj_ws = value

    @property
    def bogus(self) -> Optional[str]:
        """
        Xxx
        """
        return self.__bogus

    @bogus.setter
    def bogus(self, value: Optional[str]) -> None:
        """
        xxx
        """
        if value != self.__bogus:
            self.__log_read_write_value("bogus", value)
            self.__bogus = value

    @property
    def removed_chars_at_start_of_line(self) -> Optional[int]:
        """
        Xxx
        """
        return self.__removed_chars_at_start_of_line

    @removed_chars_at_start_of_line.setter
    def removed_chars_at_start_of_line(self, value: Optional[int]) -> None:
        """
        xxx
        """
        if value != self.__removed_chars_at_start_of_line:
            self.__log_read_write_value("removed_chars_at_start_of_line", value)
            self.__removed_chars_at_start_of_line = value

    @property
    def block_quote_data(self) -> BlockQuoteData:
        """
        Xxx
        """
        return self.__block_quote_data

    @block_quote_data.setter
    def block_quote_data(self, value: BlockQuoteData) -> None:
        """
        xxx
        """
        if value != self.__block_quote_data:
            self.__log_read_write_value("block_quote_data", value)
            self.__block_quote_data = value

    @property
    def end_container_indices(self) -> ContainerIndices:
        """
        Xxx
        """
        return self.__end_container_indices

    @end_container_indices.setter
    def end_container_indices(self, value: ContainerIndices) -> None:
        """
        xxx
        """
        if value != self.__end_container_indices:
            self.__log_read_write_value("end_container_indices", value)
            self.__end_container_indices = value

    @property
    def requeue_line_info(self) -> Optional[RequeueLineInfo]:
        """
        Xxx
        """
        return self.__requeue_line_info

    @requeue_line_info.setter
    def requeue_line_info(self, value: RequeueLineInfo) -> None:
        """
        xxx
        """
        if value != self.__requeue_line_info:
            self.__log_read_write_value("requeue_line_info", value)
            self.__requeue_line_info = value

    @property
    def current_container_blocks(self) -> List[StackToken]:
        """
        Xxx
        """
        return self.__current_container_blocks

    @current_container_blocks.setter
    def current_container_blocks(self, value: List[StackToken]) -> None:
        """
        xxx
        """
        if value != self.__current_container_blocks:
            self.__log_read_write_value("current_container_blocks", value)
            self.__current_container_blocks = value

    def get_recurse_fn(
        self,
    ) -> ParseForContainerBlocksProtocol:
        """
        xxx
        """
        return self.__recurse_fn

    def is_leaf_tokens_empty(self) -> bool:
        """
        xxx
        """
        return not self.__leaf_tokens

    def extend_leaf_tokens(self, tokens_to_append: List[MarkdownToken]) -> None:
        """
        xxx
        """
        self.__leaf_tokens.extend(tokens_to_append)
        self.__log_read_write_value("leaf_tokens", self.__leaf_tokens)

    def clear_container_tokens(self) -> None:
        """
        xxx
        """
        self.__container_level_tokens.clear()
        self.__log_read_write_value(
            "container_level_tokens", self.__container_level_tokens
        )

    def extend_container_tokens(self, tokens_to_append: List[MarkdownToken]) -> None:
        """
        xxx
        """
        self.__container_level_tokens.extend(tokens_to_append)
        self.__log_read_write_value(
            "container_level_tokens", self.__container_level_tokens
        )

    def extend_container_tokens_with_leaf_tokens(self) -> None:
        """
        xxx
        """
        self.extend_container_tokens(self.__leaf_tokens)
        self.__leaf_tokens.clear()

    @property
    def container_tokens(self) -> List[MarkdownToken]:
        """
        Xxx
        """
        return self.__container_level_tokens

    @property
    def leaf_tokens(self) -> List[MarkdownToken]:
        """
        Xxx
        """
        return self.__leaf_tokens


# pylint: enable=too-many-public-methods
# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that ensures that nested Unordered List Items
start at predictable positions.
"""

from typing import List, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.utils.container_token_manager import ContainerTokenManager
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd007(RulePlugin):
    """
    Class to implement a plugin that ensures that nested Unordered List Items
    start at predictable positions.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__indent_basis = 0
        self.__start_indented = False
        self.__container_manager = ContainerTokenManager()

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="ul-indent",
            plugin_id="MD007",
            plugin_enabled_by_default=True,
            plugin_description="Unordered list indentation",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md007.md",
            plugin_configuration="indent,start_indented",
            plugin_supports_fix=True,
            plugin_fix_level=3,
        )

    @classmethod
    def __validate_configuration_indent(cls, found_value: int) -> None:
        if found_value < 2 or found_value > 4:
            raise ValueError("Allowable values are between 2 and 4.")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__indent_basis = (
            self.plugin_configuration.get_integer_property_with_default(
                "indent",
                2,
                valid_value_fn=self.__validate_configuration_indent,
            )
        )
        self.__start_indented = (
            self.plugin_configuration.get_boolean_property_with_default(
                "start_indented",
                False,
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("indent", self.__indent_basis),
            QueryConfigItem("start_indented", self.__start_indented),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__container_manager.clear()

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        # print(f">>>{token}".replace(ParserHelper.newline_character, "\\n"))
        self.__container_manager.premanage_container_tokens(token)

        if token.is_unordered_list_start or (
            token.is_new_list_item
            and self.__container_manager.container_token_stack[
                -1
            ].is_unordered_list_start
        ):
            self.__check(context, token)

        self.__container_manager.manage_container_tokens(token)

    def __calculate_base_column_ordered_list(
        self, stack_index: int, ignore_list_starts: bool, container_base_column: int
    ) -> Tuple[bool, int]:
        if not ignore_list_starts:
            list_token = cast(
                ListStartMarkdownToken,
                self.__container_manager.container_token_stack[stack_index],
            )
            current_indent_level = (
                list_token.last_new_list_token.indent_level
                if list_token.last_new_list_token is not None
                else list_token.indent_level
            )
            container_base_column += current_indent_level
        ignore_list_starts = True
        return ignore_list_starts, container_base_column

    def __calculate_base_column_block_quote(
        self, stack_index: int, container_base_column: int, block_quote_base: int
    ) -> Tuple[bool, int, int]:
        block_quote_token = cast(
            BlockQuoteMarkdownToken,
            self.__container_manager.container_token_stack[stack_index],
        )
        bq_index = self.__container_manager.bq_line_index[stack_index + 1]
        assert block_quote_token.bleading_spaces is not None
        split_leading_spaces = block_quote_token.bleading_spaces.split(
            ParserHelper.newline_character
        )
        # print(f"bq_index={bq_index},split_leading_spaces={split_leading_spaces}")
        # print(f"split_leading_spaces[bq_index]={split_leading_spaces[bq_index]}=")
        if not block_quote_base:
            block_quote_base = container_base_column + len(
                split_leading_spaces[bq_index]
            )
        container_base_column += len(split_leading_spaces[bq_index])
        ignore_list_starts = False
        return ignore_list_starts, container_base_column, block_quote_base

    def __calculate_base_column(self) -> Tuple[int, int, int]:
        container_base_column = 0
        block_quote_base = 0
        list_depth = 0
        if self.__container_manager.container_token_stack:
            stack_index = len(self.__container_manager.container_token_stack) - 1
            while (
                stack_index >= 0
                and self.__container_manager.container_token_stack[
                    stack_index
                ].is_unordered_list_start
            ):
                list_depth += 1
                stack_index -= 1
            # print(f"stack_index={stack_index}")
            ignore_list_starts = False
            while stack_index >= 0:
                # print(f"stack_index>{stack_index}," + \
                #   f"token={self.__container_token_stack[stack_index]}".replace(ParserHelper.newline_character, "\\n"))
                if self.__container_manager.container_token_stack[
                    stack_index
                ].is_ordered_list_start:
                    (
                        ignore_list_starts,
                        container_base_column,
                    ) = self.__calculate_base_column_ordered_list(
                        stack_index, ignore_list_starts, container_base_column
                    )
                elif self.__container_manager.container_token_stack[
                    stack_index
                ].is_block_quote_start:
                    (
                        ignore_list_starts,
                        container_base_column,
                        block_quote_base,
                    ) = self.__calculate_base_column_block_quote(
                        stack_index, container_base_column, block_quote_base
                    )
                # print(f"container_base_column>{container_base_column}")
                stack_index -= 1
        return container_base_column, block_quote_base, list_depth

    def __check_apply_fix(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        adjusted_column_number: int,
        calculated_column_number: int,
    ) -> None:
        list_token = cast(ListStartMarkdownToken, token)

        # column_delta is the space before the list start and follow_space_delta
        # is the space following the list start, before the list text commences.
        column_delta = adjusted_column_number - calculated_column_number
        follow_space_delta = (list_token.indent_level - list_token.column_number) - 1

        assert len(list_token.extracted_whitespace) >= column_delta
        adjusted_whitespace = list_token.extracted_whitespace[:-column_delta]

        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "extracted_whitespace",
            adjusted_whitespace,
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "indent_level",
            list_token.indent_level - column_delta - follow_space_delta,
        )
        if not token.is_new_list_item:
            self.register_fix_token_request(
                context,
                token,
                "next_token",
                "column_number",
                list_token.column_number - column_delta,
            )
            if list_token.leading_spaces:
                new_spaces = []
                total_delta = column_delta + follow_space_delta
                for i in list_token.leading_spaces.split("\n"):
                    if len(i) >= list_token.indent_level:
                        i = i[:-total_delta]
                    new_spaces.append(i)
                self.register_fix_token_request(
                    context,
                    list_token,
                    "next_token",
                    "leading_spaces",
                    "\n".join(new_spaces),
                )

    def __check(self, context: PluginScanContext, token: MarkdownToken) -> None:
        # print(f"{token}".replace(ParserHelper.newline_character, "\\n"))
        # print(f"{self.__container_token_stack}".replace(ParserHelper.newline_character, "\\n"))
        # print(f"{self.__bq_line_index}".replace(ParserHelper.newline_character, "\\n"))

        (
            container_base_column,
            block_quote_base,
            list_depth,
        ) = self.__calculate_base_column()
        # print(f"container_base_column={container_base_column}, block_quote_base={block_quote_base}, list_depth={list_depth}")

        # print(f"list_depth={list_depth}")
        if token.is_new_list_item:
            # print(f"list_depth={list_depth}")
            list_depth -= 1

        if self.__start_indented:
            # print(f"list_depth={list_depth}")
            list_depth += 1

        adjusted_column_number = token.column_number - 1 - container_base_column
        # print(f"adjusted_column_number={adjusted_column_number} = token.column_number({token.column_number}) -1 -container_base_column={container_base_column}")
        calculated_column_number = list_depth * self.__indent_basis
        # print(f"adjusted_column_number={adjusted_column_number}, calculated_column_number=" + \
        #   f"{calculated_column_number},block_quote_base={block_quote_base}")
        if adjusted_column_number > calculated_column_number:
            if context.in_fix_mode:
                self.__check_apply_fix(
                    context, token, adjusted_column_number, calculated_column_number
                )
            else:
                # print(f"container_base_column={container_base_column}")
                if block_quote_base:
                    container_base_column -= block_quote_base
                elif container_base_column:
                    container_base_column += 1
                extra_error_information = (
                    f"Expected: {calculated_column_number + container_base_column}, "
                    + f"Actual={adjusted_column_number + container_base_column}"
                )
                self.report_next_token_error(
                    context, token, extra_error_information=extra_error_information
                )



"""
Module to provide for a simplified way to setup logging.
"""

import argparse
import logging
import sys
from typing import List, Optional, Tuple

from application_properties import ApplicationProperties

LOGGER = logging.getLogger(__name__)


class ApplicationLogging:
    """
    Class to provide for a simplified way to setup logging.
    """

    log_level_critical = "CRITICAL"
    log_level_error = "ERROR"
    log_level_warning = "WARNING"
    log_level_info = "INFO"
    log_level_debug = "DEBUG"
    __available_log_maps = {
        log_level_critical: logging.CRITICAL,
        log_level_error: logging.ERROR,
        log_level_warning: logging.WARNING,
        log_level_info: logging.INFO,
        log_level_debug: logging.DEBUG,
    }
    __log_level_to_name_map = {
        logging.CRITICAL: log_level_critical,
        logging.ERROR: log_level_error,
        logging.WARNING: log_level_warning,
        logging.INFO: log_level_info,
        logging.DEBUG: log_level_debug,
    }

    def __init__(
        self,
        application_properties: Optional[ApplicationProperties] = None,
        show_stack_trace: bool = False,
        default_log_level: str = log_level_warning,
    ):
        self.__show_stack_trace = show_stack_trace
        self.__default_log_level = default_log_level
        self.__properties = application_properties
        self.__new_handler: Optional[logging.FileHandler] = None

    def pre_initialize_with_args(self, args: argparse.Namespace) -> None:
        """
        Pre-Initialize the logging for the application.  This is done to provide
        logging before any configuration has been initialized and sets logging
        to default values.
        """
        self.pre_initialize(args.show_stack_trace)

    def pre_initialize(self, show_stack_trace: bool = False) -> None:
        """
        Pre-Initialize the logging for the application.  This is done to provide
        logging before any configuration has been initialized and sets logging
        to default values.
        """
        self.__show_stack_trace = show_stack_trace
        base_logger = logging.getLogger()
        new_level = logging.DEBUG if show_stack_trace else logging.WARNING
        base_logger.setLevel(new_level)
        if show_stack_trace:
            base_logger.warning(
                "Application logging set to '%s'.",
                logging.getLevelName(base_logger.level),
            )
        logging.basicConfig(stream=sys.stdout, level=new_level)

    def __calculate_effective_levels(
        self, args: argparse.Namespace
    ) -> Tuple[str, Optional[str]]:
        effective_log_file = args.log_file
        if not effective_log_file and self.__properties:
            effective_log_file = self.__properties.get_string_property("log.file")

        effective_log_level = args.log_level
        if not effective_log_level and self.__properties:
            effective_log_level = self.__properties.get_string_property(
                "log.level",
                valid_value_fn=ApplicationLogging.validate_log_level_type,
            )
        if effective_log_level is None:
            effective_log_level = self.__default_log_level

        return effective_log_level, effective_log_file

    def initialize(self, args: argparse.Namespace) -> None:
        """
        Initialize the logging for the application.
        """

        try:
            effective_log_level, effective_log_file = self.__calculate_effective_levels(
                args
            )

            if effective_log_file:
                self.__new_handler = logging.FileHandler(effective_log_file)
                formatter = logging.Formatter("%(levelname)s %(asctime)s %(message)s")
                self.__new_handler.setFormatter(formatter)
                logging.getLogger().addHandler(self.__new_handler)
            else:
                temp_log_level = (
                    logging.DEBUG if self.__show_stack_trace else logging.CRITICAL
                )
                logging.basicConfig(stream=sys.stdout, level=temp_log_level)

            log_level_to_enact = ApplicationLogging.__available_log_maps[
                effective_log_level
            ]

            logging.getLogger().setLevel(log_level_to_enact)
        except ValueError:
            raise
        except Exception as this_exception:
            self.terminate()
            raise ApplicationLoggingException(
                "Failure initializing logging subsystem."
            ) from this_exception

    def terminate(self) -> None:
        """
        Terminate the logging for the application.
        """
        if self.__new_handler:
            self.__new_handler.close()
            self.__new_handler = None

    @staticmethod
    def add_default_command_line_arguments(
        parser_to_add_to: argparse.ArgumentParser,
    ) -> None:
        """
        Add a set of default command line arguments to an argparse styled command line.
        """
        parser_to_add_to.add_argument(
            "--log-level",
            dest="log_level",
            action="store",
            help="minimum level required to log messages",
            type=ApplicationLogging.validate_log_level_type,
            choices=list(ApplicationLogging.__available_log_maps.keys()),
        )
        parser_to_add_to.add_argument(
            "--log-file",
            dest="log_file",
            action="store",
            help="destination file for log messages",
        )

    @staticmethod
    def get_valid_log_levels() -> List[str]:
        """
        Return a sorted list of the available log levels as strings.
        """
        return list(ApplicationLogging.__available_log_maps.keys())

    @staticmethod
    def is_valid_log_level_type(argument: str) -> bool:
        """
        Check to see if the supplied arguments is a valid log level.
        """
        try:
            ApplicationLogging.validate_log_level_type(argument)
            return True
        except ValueError:
            return False

    @staticmethod
    def validate_log_level_type(argument: str) -> str:
        """
        Function to help argparse limit the valid log levels.
        """
        if argument in ApplicationLogging.__available_log_maps:
            return argument
        raise ValueError(f"Value '{argument}' is not a valid log level.")

    @staticmethod
    def translate_log_level(value: str) -> int:
        """
        Translate a log string, such as "DEBUG" into the logging system's
        numeric values.
        """
        if value not in ApplicationLogging.__available_log_maps:
            return logging.NOTSET
        return ApplicationLogging.__available_log_maps[value]


class ApplicationLoggingException(Exception):
    """
    Class to provide for a specific
    """



"""
Module to provide for an encapsulation of a plugin that was discovered.
"""

# pylint: disable=too-many-instance-attributes
from dataclasses import dataclass
from typing import List, Optional

from pymarkdown.plugin_manager.rule_plugin import RulePlugin


@dataclass(frozen=True)
class FoundPlugin:
    """
    Encapsulation of a plugin that was discovered.  While similar to the PluginDetails
    class, this is meant for an internal representation of the plugin, and not the
    external information provided.
    """

    plugin_id: str
    plugin_names: List[str]
    plugin_description: str
    plugin_instance: RulePlugin
    plugin_enabled_by_default: bool
    plugin_version: str
    plugin_interface_version: int
    plugin_file_name: str
    plugin_url: Optional[str]
    plugin_configuration: Optional[str]
    plugin_supports_fix: bool
    plugin_fix_level: int
    plugin_identifiers: List[str]


# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that looks for spaces within emphasis sections.
"""

from dataclasses import dataclass
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


@dataclass
class EligibleEmphasis:
    """
    Emphasis characters that are eligible for actual emphasis.
    """

    emphasis_character: str
    start_index: int
    found_length: int
    character_before: Optional[str] = None
    character_after: Optional[str] = None
    text_token: Optional[TextMarkdownToken] = None


@dataclass
class PendingFixes:
    """
    Fixes that are pending being applied to the token.
    """

    token_to_modify: TextMarkdownToken
    start_index: int
    end_index: int


class RuleMd037(RulePlugin):
    """
    Class to implement a plugin that looks for spaces within emphasis sections.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__pending_fixes: List[PendingFixes] = []
        self.__block_token: Optional[MarkdownToken] = None
        self.__past_emphasis_list: List[EligibleEmphasis] = []

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="no-space-in-emphasis",
            plugin_id="MD037",
            plugin_enabled_by_default=True,
            plugin_description="Spaces inside emphasis markers",
            plugin_version="0.5.2",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md037.md",
            plugin_supports_fix=True,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__block_token = None
        self.__past_emphasis_list = []

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_paragraph or token.is_setext_heading or token.is_atx_heading:
            self.__block_token = token
            self.__past_emphasis_list = []
        elif (
            token.is_paragraph_end
            or token.is_setext_heading_end
            or token.is_atx_heading_end
        ):
            self.__block_token = None
            self.__past_emphasis_list = []
            if self.__pending_fixes:
                self.__process_fixes(context)
        elif (
            token.is_text
            and self.__block_token is not None
            and (
                self.__block_token.is_paragraph
                or self.__block_token.is_setext_heading
                or self.__block_token.is_atx_heading
            )
        ):
            text_token = cast(TextMarkdownToken, token)
            self.__check_text_token(context, text_token)

    def __check_text_token(
        self, context: PluginScanContext, text_token: TextMarkdownToken
    ) -> None:
        start_index = 0
        next_index: Optional[int] = -1
        while next_index is not None:
            new_emphasis, next_index = self.__find_next_eligible_emphasis(
                text_token, start_index
            )
            if new_emphasis is not None:
                if (
                    self.__past_emphasis_list
                    and self.__past_emphasis_list[-1].emphasis_character
                    == new_emphasis.emphasis_character
                    and self.__past_emphasis_list[-1].found_length
                    == new_emphasis.found_length
                ):
                    self.__check(context, self.__past_emphasis_list[-1], new_emphasis)
                    del self.__past_emphasis_list[-1]
                else:
                    self.__past_emphasis_list.append(new_emphasis)
            if next_index is not None:
                found_emphasis_length = (
                    new_emphasis.found_length if new_emphasis is not None else 1
                )
                start_index = next_index + 1 + found_emphasis_length

    def __check(
        self,
        context: PluginScanContext,
        eligible_before: EligibleEmphasis,
        eligible_after: EligibleEmphasis,
    ) -> None:
        before_surrounded = (
            eligible_before.character_before == " "
            and eligible_before.character_after == " "
        )
        after_surrounded = (
            eligible_after.character_before == " "
            and eligible_after.character_after == " "
        )

        if before_surrounded and eligible_after.character_after == " ":
            if context.in_fix_mode:
                assert eligible_before.text_token is not None
                self.__fix(eligible_before.text_token, eligible_before, True)
            else:
                self.__report(context, eligible_before, eligible_after.found_length)

        if after_surrounded and eligible_before.character_before == " ":
            if context.in_fix_mode:
                assert eligible_after.text_token is not None
                self.__fix(eligible_after.text_token, eligible_after, False)
            else:
                self.__report(context, eligible_after, -1)

    def __process_fixes(self, context: PluginScanContext) -> None:
        current_token = None
        current_text = ""
        current_delta = 0
        for new_bob in self.__pending_fixes:
            if current_token != new_bob.token_to_modify:
                if current_token is not None:
                    self.register_fix_token_request(
                        context, current_token, "next_token", "token_text", current_text
                    )
                current_token = new_bob.token_to_modify
                current_text = current_token.token_text
                current_delta = 0
            current_text = (
                current_text[: new_bob.start_index - current_delta]
                + current_text[new_bob.end_index - current_delta :]
            )
            current_delta += new_bob.end_index - new_bob.start_index
        assert current_token is not None
        self.register_fix_token_request(
            context, current_token, "next_token", "token_text", current_text
        )
        self.__pending_fixes.clear()

    def __fix(
        self,
        token: TextMarkdownToken,
        eligible_after: EligibleEmphasis,
        was_after: bool,
    ) -> None:
        if was_after:
            start_fix_index = eligible_after.start_index + eligible_after.found_length
            end_fix_index, _ = ParserHelper.collect_while_one_of_characters_verified(
                token.token_text, start_fix_index, " \t"
            )
        else:
            _, start_fix_index = (
                ParserHelper.collect_backwards_while_one_of_characters_verified(
                    token.token_text, eligible_after.start_index - 1, " \t"
                )
            )
            end_fix_index = eligible_after.start_index
        self.__pending_fixes.append(PendingFixes(token, start_fix_index, end_fix_index))

    def __report(
        self,
        context: PluginScanContext,
        elibible_before: EligibleEmphasis,
        column_adjust: int,
    ) -> None:
        assert elibible_before.text_token is not None
        before_eligible_text = elibible_before.text_token.token_text[
            : elibible_before.start_index
        ]
        adj_before_eligible_text = ParserHelper.remove_all_from_text(
            before_eligible_text
        )
        line_number_delta = ParserHelper.count_newlines_in_text(
            adj_before_eligible_text
        )
        if line_number_delta:
            last_newline_index = adj_before_eligible_text.rindex("\n") + 1
            column_number_delta = len(adj_before_eligible_text) - last_newline_index
        else:
            column_number_delta = len(adj_before_eligible_text)
        self.report_next_token_error(
            context,
            elibible_before.text_token,
            line_number_delta=line_number_delta
            + context.calc_pragma_offset(elibible_before.text_token, line_number_delta),
            column_number_delta=(column_number_delta + column_adjust),
        )

    def __find_next_eligible_emphasis(
        self, text_token: TextMarkdownToken, start_index: int
    ) -> Tuple[Optional[EligibleEmphasis], Optional[int]]:

        asterisk_index = text_token.token_text.find("*", start_index)
        underscore_index = text_token.token_text.find("_", start_index)
        if asterisk_index == -1 and underscore_index == -1:
            return None, None

        if (
            asterisk_index != -1
            and underscore_index != -1
            and asterisk_index < underscore_index
            or asterisk_index != -1
            and underscore_index == -1
        ):
            start_index = asterisk_index
            emphasis_character = "*"
        else:
            start_index = underscore_index
            emphasis_character = "_"

        character_before = (
            text_token.token_text[start_index - 1] if start_index else None
        )
        if character_before == "\b":
            return None, start_index
        found_length, _ = ParserHelper.collect_while_character_verified(
            text_token.token_text, start_index, emphasis_character
        )
        after_index = start_index + found_length
        character_after = (
            text_token.token_text[after_index]
            if after_index < len(text_token.token_text)
            else None
        )
        if character_before == "\a" and character_after == "\a":
            return None, start_index
        is_before = (
            ParserHelper.is_character_at_index_whitespace(character_before, 0)
            if character_before
            else False
        )
        is_after = (
            ParserHelper.is_character_at_index_whitespace(character_after, 0)
            if character_after
            else False
        )
        if is_before or is_after:
            return (
                EligibleEmphasis(
                    emphasis_character,
                    start_index,
                    found_length,
                    character_before,
                    character_after,
                    text_token,
                ),
                start_index,
            )
        return None, start_index



"""
Module to implement a plugin that ensures that all Unordered List Items
start at the beginning of the line.
"""

from typing import List, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd006(RulePlugin):
    """
    Class to implement a plugin that ensures that all Unordered List Items
    start at the beginning of the line.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__token_stack: List[MarkdownToken] = []

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="ul-start-left",
            plugin_id="MD006",
            plugin_enabled_by_default=False,
            plugin_description="Consider starting bulleted lists at the beginning of the line",
            plugin_version="0.5.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md006.md",
            plugin_supports_fix=True,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__token_stack = []

    def __calculate_expected_indent(self) -> int:
        expected_indent = 0
        if len(self.__token_stack) > 1:
            if self.__token_stack[-2].is_block_quote_start:
                block_quote_token = cast(
                    BlockQuoteMarkdownToken, self.__token_stack[-2]
                )
                assert block_quote_token.bleading_spaces is not None
                split_spaces = block_quote_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
                expected_indent = len(split_spaces[0]) + (
                    block_quote_token.column_number - 1
                )
            else:
                list_token = cast(ListStartMarkdownToken, self.__token_stack[-2])
                expected_indent = list_token.indent_level
        return expected_indent

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_list_start or token.is_block_quote_start:
            self.__token_stack.append(token)
            if token.is_unordered_list_start:
                expected_indent = self.__calculate_expected_indent()
                if delta := (1 + expected_indent) - token.column_number:
                    self.__report_or_fix(context, token, delta)
        elif token.is_list_end or token.is_block_quote_end:
            del self.__token_stack[-1]
        elif token.is_new_list_item:
            if self.__token_stack[-1].is_unordered_list_start:
                expected_indent = self.__calculate_expected_indent()
                if delta := (1 + expected_indent) - token.column_number:
                    self.__report_or_fix(context, token, delta)

    def __report_or_fix(
        self, context: PluginScanContext, token: MarkdownToken, adjust_amount: int
    ) -> None:
        if context.in_fix_mode:
            list_start_token = cast(ListStartMarkdownToken, token)
            self.register_fix_token_request(
                context,
                token,
                "next_token",
                "indent_level",
                list_start_token.indent_level + adjust_amount,
            )
            if not token.is_new_list_item:
                self.register_fix_token_request(
                    context, token, "next_token", "extracted_whitespace", ""
                )
                self.register_fix_token_request(
                    context,
                    token,
                    "next_token",
                    "column_number",
                    list_start_token.column_number + adjust_amount,
                )
            else:
                self.register_fix_token_request(
                    context,
                    token,
                    "next_token",
                    "extracted_whitespace",
                    list_start_token.extracted_whitespace[:adjust_amount],
                )
        else:
            self.report_next_token_error(context, token)



"""
Inline processing
"""

import logging
from typing import List, cast

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.extension_manager.extension_manager import ExtensionManager
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.inline_handler_helper import InlineHandlerHelper
from pymarkdown.inline.inline_helper import InlineHelper
from pymarkdown.inline.inline_text_block_helper import InlineTextBlockHelper
from pymarkdown.plugins.utils.leading_space_index_tracker import (
    LeadingSpaceIndexTracker,
)
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class InlineProcessor:
    """
    Handle the inline processing of the token stream.
    """

    @staticmethod
    def initialize(extension_manager: ExtensionManager) -> None:
        """
        Initialize the inline processor subsystem.
        """
        InlineHandlerHelper.initialize(extension_manager)

    @staticmethod
    def __parse_inline_tracker_start(
        lsi_tracker: LeadingSpaceIndexTracker, token: MarkdownToken
    ) -> None:
        if token.is_block_quote_start or token.is_list_start:
            lsi_tracker.open_container(token)
        # if token.is_block_quote_end or token.is_list_end:
        assert not (token.is_block_quote_end or token.is_list_end)
        # lsi_tracker.register_container_end(token)
        # elif
        lsi_tracker.track_since_last_non_end_token(token)

    @staticmethod
    def __parse_inline_tracker_per_token(
        lsi_tracker: LeadingSpaceIndexTracker, token: MarkdownToken
    ) -> None:
        if not token.is_end_token or token.is_end_of_stream:
            while lsi_tracker.have_any_registered_container_ends():
                lsi_tracker.process_container_end(token)
        if token.is_block_quote_start or token.is_list_start:
            lsi_tracker.open_container(token)
        elif token.is_block_quote_end or token.is_list_end:
            lsi_tracker.register_container_end(token)
        lsi_tracker.track_since_last_non_end_token(token)

    @staticmethod
    def parse_inline(
        coalesced_results: List[MarkdownToken],
        parse_properties: ParseBlockPassProperties,
    ) -> List[MarkdownToken]:
        """
        Parse and resolve any inline elements.
        """
        POGGER.info("coalesced_results")
        POGGER.info("-----")
        for next_token in coalesced_results:
            POGGER.info(">>$<<", next_token)
        POGGER.info("-----")

        coalesced_stack: List[MarkdownToken] = []
        current_token = coalesced_results[0]
        coalesced_list: List[MarkdownToken] = []
        coalesced_list.extend(coalesced_results[:1])

        POGGER.debug("STACK?:$", current_token)
        if current_token.is_container:
            POGGER.debug("STACK:$", coalesced_stack)
            coalesced_stack.append(current_token)
            POGGER.debug("STACK-ADD:$", current_token)
            POGGER.debug("STACK:$", coalesced_stack)
            if current_token.is_block_quote_start:
                block_quote_token = cast(BlockQuoteMarkdownToken, current_token)
                block_quote_token.leading_text_index = 0
                POGGER.info("-->last->block->$", block_quote_token.leading_text_index)
            else:
                POGGER.info("-->not bq-")

        token = coalesced_results[0]
        lsi_tracker = LeadingSpaceIndexTracker()
        InlineProcessor.__parse_inline_tracker_start(lsi_tracker, token)
        for coalesce_index in range(1, len(coalesced_results)):

            token = coalesced_results[coalesce_index]
            InlineProcessor.__parse_inline_tracker_per_token(lsi_tracker, token)

            InlineProcessor.__process_next_coalesce_item(
                coalesced_results,
                coalesce_index,
                coalesced_list,
                coalesced_stack,
                parse_properties,
                lsi_tracker,
            )
        return coalesced_list

    @staticmethod
    def __adjust_stack(
        coalesced_results: List[MarkdownToken],
        coalesced_stack: List[MarkdownToken],
        coalesce_index: int,
    ) -> None:
        current_token = coalesced_results[coalesce_index]
        POGGER.debug("STACK?:$", current_token)
        if current_token.is_container:
            if current_token.is_new_list_item:
                assert coalesced_stack[
                    -1
                ].is_list_start, "New list items can only exist within lists."
                list_token = cast(ListStartMarkdownToken, coalesced_stack[-1])
                new_item_token = cast(NewListItemMarkdownToken, current_token)
                list_token.adjust_for_new_list_item(
                    new_item_token, skip_adjustment=True
                )
            else:
                POGGER.debug("STACK:$", coalesced_stack)
                coalesced_stack.append(current_token)
                POGGER.debug("STACK-ADD:$", current_token)
                POGGER.debug("STACK:$", coalesced_stack)
                if current_token.is_block_quote_start:
                    block_quote_token = cast(BlockQuoteMarkdownToken, current_token)
                    block_quote_token.leading_text_index = 0
                    POGGER.info(
                        "-->last->block->$", block_quote_token.leading_text_index
                    )
                else:
                    list_token = cast(ListStartMarkdownToken, current_token)
                    assert (
                        list_token.last_new_list_token is None
                    ), "Cannot have any new list items registered."
                    POGGER.info("-->not bq-")

        elif current_token.is_list_end or current_token.is_block_quote_end:
            POGGER.debug("STACK:$", coalesced_stack)
            del coalesced_stack[-1]
            POGGER.debug(
                "STACK-REMOVE:$",
                current_token,
            )
            POGGER.debug("STACK:$", coalesced_stack)

    @staticmethod
    def __process_next_coalesce_item(
        coalesced_results: List[MarkdownToken],
        coalesce_index: int,
        coalesced_list: List[MarkdownToken],
        coalesced_stack: List[MarkdownToken],
        parse_properties: ParseBlockPassProperties,
        lsi_tracker: LeadingSpaceIndexTracker,
    ) -> None:
        POGGER.info("coalesced_results:$<", coalesced_list[-1])
        POGGER.info("coalesced_stack:$<", coalesced_stack)
        for i in range(len(coalesced_stack) - 1, -1, -1):
            if coalesced_stack[i].is_block_quote_start:
                block_quote_token = cast(BlockQuoteMarkdownToken, coalesced_stack[i])
                POGGER.info(
                    "$-->last->block->$", i, block_quote_token.leading_text_index
                )

        InlineProcessor.__process_next_coalesce_item_inner(
            coalesced_list,
            coalesced_results,
            coalesce_index,
            coalesced_stack,
            parse_properties,
            lsi_tracker,
        )

        InlineProcessor.__adjust_stack(
            coalesced_results, coalesced_stack, coalesce_index
        )

    @staticmethod
    def __process_next_coalesce_item_inner(
        coalesced_list: List[MarkdownToken],
        coalesced_results: List[MarkdownToken],
        coalesce_index: int,
        coalesced_stack: List[MarkdownToken],
        parse_properties: ParseBlockPassProperties,
        lsi_tracker: LeadingSpaceIndexTracker,
    ) -> None:
        if coalesced_results[coalesce_index].is_text and (
            coalesced_list[-1].is_paragraph
            or coalesced_list[-1].is_setext_heading
            or coalesced_list[-1].is_atx_heading
            or coalesced_list[-1].is_code_block
            or coalesced_list[-1].is_table_header_item
            or coalesced_list[-1].is_table_row_item
        ):
            if coalesced_list[-1].is_code_block:
                processed_tokens = InlineProcessor.__parse_code_block(
                    coalesced_results,
                    coalesce_index,
                    coalesced_list,
                    coalesced_stack,
                    lsi_tracker,
                )
            elif coalesced_list[-1].is_setext_heading:
                processed_tokens = InlineProcessor.__parse_setext_heading(
                    coalesced_results, coalesce_index, coalesced_stack, parse_properties
                )
            elif coalesced_list[-1].is_atx_heading:
                processed_tokens = InlineProcessor.__parse_atx_heading(
                    coalesced_results,
                    coalesce_index,
                    coalesced_stack,
                    coalesced_list,
                    parse_properties,
                )
            elif (
                coalesced_list[-1].is_table_header_item
                or coalesced_list[-1].is_table_row_item
            ):
                processed_tokens = InlineProcessor.__parse_table_header_item(
                    coalesced_results,
                    coalesce_index,
                    coalesced_stack,
                    coalesced_list,
                    parse_properties,
                )
            else:
                processed_tokens = InlineProcessor.__parse_paragraph(
                    coalesced_list,
                    coalesced_results,
                    coalesce_index,
                    coalesced_stack,
                    parse_properties,
                )
            coalesced_list.extend(processed_tokens)
        else:
            coalesced_list.append(coalesced_results[coalesce_index])

    @staticmethod
    def __parse_paragraph(
        coalesced_list: List[MarkdownToken],
        coalesced_results: List[MarkdownToken],
        coalesce_index: int,
        coalesced_stack: List[MarkdownToken],
        parse_properties: ParseBlockPassProperties,
    ) -> List[MarkdownToken]:
        assert coalesced_list[
            -1
        ].is_paragraph, "Must already be within a paragraph block."
        paragraph_token = cast(ParagraphMarkdownToken, coalesced_list[-1])
        assert coalesced_results[
            coalesce_index
        ].is_text, "Results must already be text tokens."
        text_token = cast(TextMarkdownToken, coalesced_results[coalesce_index])
        POGGER.debug(
            ">>before_add_ws>>$>>add>>$>>",
            coalesced_list[-1],
            text_token.extracted_whitespace,
        )
        paragraph_token.add_whitespace(text_token.extracted_whitespace)
        POGGER.debug(">>after_add_ws>>$", coalesced_list[-1])
        POGGER.debug(
            ">>text_token>>$", ParserHelper.make_whitespace_visible(str(text_token))
        )
        POGGER.debug(
            "text_token.token_text>:$:<",
            ParserHelper.make_whitespace_visible(str(text_token.token_text)),
        )
        return InlineTextBlockHelper.process_inline_text_block(
            parse_properties,
            text_token.token_text,
            coalesced_stack,
            is_para=True,
            para_space=text_token.extracted_whitespace,
            line_number=text_token.line_number,
            column_number=text_token.column_number,
            para_owner=paragraph_token,
            tabified_text=text_token.tabified_text,
        )

    @staticmethod
    def __parse_atx_heading(
        coalesced_results: List[MarkdownToken],
        coalesce_index: int,
        coalesced_stack: List[MarkdownToken],
        coalesced_list: List[MarkdownToken],
        parse_properties: ParseBlockPassProperties,
    ) -> List[MarkdownToken]:
        assert coalesced_results[
            coalesce_index
        ].is_text, "Coalesced tokens must be text."
        text_token = cast(TextMarkdownToken, coalesced_results[coalesce_index])
        POGGER.debug("atx-block>>$<<", text_token)
        POGGER.debug(
            "atx-block-text>>$<<",
            text_token.token_text,
        )
        POGGER.debug(
            "atx-block-ws>>$<<",
            text_token.extracted_whitespace,
        )

        assert coalesced_list[
            -1
        ].is_atx_heading, "Final coalseced token must be an ATX token."
        atx_token = cast(AtxHeadingMarkdownToken, coalesced_list[-1])
        POGGER.debug(">>text_token>>$", text_token)
        return InlineTextBlockHelper.process_inline_text_block(
            parse_properties,
            text_token.token_text,
            coalesced_stack,
            text_token.extracted_whitespace,
            line_number=text_token.line_number,
            column_number=(
                text_token.column_number
                + len(text_token.extracted_whitespace)
                + atx_token.hash_count
            ),
            tabified_text=text_token.tabified_text,
        )

    @staticmethod
    def __parse_table_header_item(
        coalesced_results: List[MarkdownToken],
        coalesce_index: int,
        coalesced_stack: List[MarkdownToken],
        coalesced_list: List[MarkdownToken],
        parse_properties: ParseBlockPassProperties,
    ) -> List[MarkdownToken]:
        assert coalesced_results[
            coalesce_index
        ].is_text, "Coalesced tokens must be text."
        text_token = cast(TextMarkdownToken, coalesced_results[coalesce_index])
        POGGER.debug("table_header>>$<<", text_token)
        POGGER.debug(
            "table_header-text>>$<<",
            text_token.token_text,
        )
        POGGER.debug(
            "table_header-ws>>$<<",
            text_token.extracted_whitespace,
        )

        # assert coalesced_list[
        #     -1
        # ].is_atx_heading, "Final coalseced token must be an ATX token."
        # atx_token = cast(AtxHeadingMarkdownToken, coalesced_list[-1])
        POGGER.debug(">>text_token>>$", text_token)
        return InlineTextBlockHelper.process_inline_text_block(
            parse_properties,
            text_token.token_text,
            coalesced_stack,
            text_token.extracted_whitespace,
            line_number=text_token.line_number,
            column_number=(
                text_token.column_number
                + len(text_token.extracted_whitespace)
                # + atx_token.hash_count
            ),
            tabified_text=text_token.tabified_text,
            is_in_table=True,
        )

    @staticmethod
    def __parse_setext_heading(
        coalesced_results: List[MarkdownToken],
        coalesce_index: int,
        coalesced_stack: List[MarkdownToken],
        parse_properties: ParseBlockPassProperties,
    ) -> List[MarkdownToken]:
        assert coalesced_results[
            coalesce_index
        ].is_text, "Coalesced tokens must be text."
        text_token = cast(TextMarkdownToken, coalesced_results[coalesce_index])
        POGGER.debug(">>text_token>>$", text_token)
        processed_tokens = InlineTextBlockHelper.process_inline_text_block(
            parse_properties,
            text_token.token_text,
            coalesced_stack,
            whitespace_to_recombine=text_token.extracted_whitespace,
            is_setext=True,
            para_space=text_token.extracted_whitespace,
            line_number=text_token.line_number,
            column_number=text_token.column_number,
            tabified_text=text_token.tabified_text,
        )
        POGGER.debug(
            "processed_tokens>>$",
            processed_tokens,
        )
        return processed_tokens

    @staticmethod
    def __parse_code_block_coalesced(
        coalesced_stack: List[MarkdownToken],
        lsi_tracker: LeadingSpaceIndexTracker,
        text_token: TextMarkdownToken,
    ) -> int:
        new_column_number = 1
        if coalesced_stack[-1].is_block_quote_start:
            block_quote_token = cast(BlockQuoteMarkdownToken, coalesced_stack[-1])
            assert block_quote_token.bleading_spaces, "Bleading spaces must be defined."
            split_leading_spaces = block_quote_token.bleading_spaces.split(
                ParserHelper.newline_character
            )
            split_index = lsi_tracker.get_tokens_block_quote_bleading_space_index(
                text_token
            )
        else:
            list_token = cast(ListStartMarkdownToken, coalesced_stack[-1])
            assert (
                list_token.leading_spaces is not None
            ), "Leading spaces must be defined."
            split_leading_spaces = list_token.leading_spaces.split(
                ParserHelper.newline_character
            )
            split_index = lsi_tracker.get_tokens_list_leading_space_index(text_token)
        if split_index < len(split_leading_spaces):
            new_column_number += len(split_leading_spaces[split_index])
        if coalesced_stack[-1].is_list_start:
            stack_index = len(coalesced_stack) - 2
            while stack_index >= 0 and coalesced_stack[stack_index].is_list_start:
                stack_index -= 1
            if stack_index >= 0:
                assert coalesced_stack[stack_index].is_block_quote_start
                block_quote_token = cast(
                    BlockQuoteMarkdownToken, coalesced_stack[stack_index]
                )
                assert (
                    block_quote_token.bleading_spaces
                ), "Bleading spaces must be defined."
                split_leading_spaces = block_quote_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
                split_index = lsi_tracker.get_tokens_list_leading_space_index(
                    text_token
                )
                new_column_number += len(split_leading_spaces[split_index])
        return new_column_number

    @staticmethod
    def __parse_code_block(
        coalesced_results: List[MarkdownToken],
        coalesce_index: int,
        coalesced_list: List[MarkdownToken],
        coalesced_stack: List[MarkdownToken],
        lsi_tracker: LeadingSpaceIndexTracker,
    ) -> List[MarkdownToken]:
        assert coalesced_results[
            coalesce_index
        ].is_text, "Coalesced tokens must be text."
        text_token = cast(TextMarkdownToken, coalesced_results[coalesce_index])
        encoded_text = InlineHelper.append_text("", text_token.token_text)
        if coalesced_list[-1].is_fenced_code_block:
            line_number_delta = text_token.line_number - coalesced_list[-1].line_number

            # POGGER.info("coalesced_stack:$<", coalesced_stack)
            if coalesced_stack:
                new_column_number = InlineProcessor.__parse_code_block_coalesced(
                    coalesced_stack, lsi_tracker, text_token
                )
            else:
                new_column_number = 1
                leading_whitespace = text_token.extracted_whitespace
                # POGGER.debug(">>$<<", text_token)
                assert (
                    ParserHelper.newline_character not in leading_whitespace
                ), "Whitespace cannot contain any newlines."
                # POGGER.info(
                #     "leading_whitespace:$<",
                #     leading_whitespace,
                # )
                leading_whitespace = ParserHelper.remove_all_from_text(
                    leading_whitespace
                )
                new_column_number = 1 + len(leading_whitespace)
        else:
            line_number_delta, new_column_number = (
                0,
                coalesced_list[-1].column_number,
            )
        processed_tokens: List[MarkdownToken] = [
            TextMarkdownToken(
                encoded_text,
                text_token.extracted_whitespace,
                line_number=coalesced_list[-1].line_number + line_number_delta,
                column_number=new_column_number,
            )
        ]
        # POGGER.debug(
        #     "new Text>>$>>",
        #     processed_tokens,
        # )
        return processed_tokens
